#ifdef DEBUG
# define allocate(n) trace_calloc(__FILE__, __LINE__, n, 1)
#endif
/*
 * Starting from seed, generate random numbers to fill an array tracep
 * of length 2^n representing a trace of the instantaneous traffic
 * generated by a self similar source with mean traffic rate M,
 * peakedness factor a and Hurst parameter H during a time interval T.
 *
 * M, a and T should be positive numbers, 0<H<1, n > 0.  M, a and H
 * are only target values for the generation.  The trace will have
 * generally different statistics, in particular the mean will
 * generally be higher than M due to the truncation of the negative
 * samples.
 *
 * The whole output vector is considered to be of time length T, so the
 *   values contained are the traffic intensities in time intervals of
 *   length T*2^-n.
 * M is the value of the traffic rate.
 * a is the peakedness factor, that is the ratio of variance to the
 *   mean of the traffic per unit time interval.
 * tracep is a pointer to a vector of floats where the output is written.
 *   If it is NULL, a memory area is allocated like this:
 *   allocate (sizeof (float) << n).
 * The return value is the address of the vector of floats filled with the
 * computed values.
 *
 * Lau, Erramilli, Wang, Willinger, "Self-similar Traffic Generation: The
 * random Midpoint Displacement Algorithm and Its Properties", ICC '95,
 * pp. 466-472.
 */

#include "libnr.h"
#ifndef NULL
# define NULL 0
#endif
#define max(a,b) ((a) > (b) ? (a) : (b))

extern int random();
extern double pow();
extern double sqrt();

float *
RMDtraffic (M, a, H, T, n, tracep)
     double M;			/* mean traffic per unit time */
     double a;			/* peakedness (var/mean per unit time) */
     double H;			/* Hurst parameter */
     double T;			/* time length of the trace */
     int n;			/* the trace length is 2^n */
     float *tracep;		/* trace pointer (array of float's) */
{
  int N = 1 << n;
  register float std;
  register float scale;
  float scalefactor = pow (2.0, -H);
  register float mean, ampl;
  register float *lastp, *sp;
  register int i, j, d;

  /* Allocate the trace array if not already done.  Set lastp to the
     last element. */
  if (tracep == NULL)
    tracep = allocate (sizeof (*tracep) << n);
  lastp = tracep + N - 1;

  std = pow (T, H);

  /* Generate a Fractional Brownian Motion process.  The first point is
     always zero.  Its proper place would be at tracep[-1], but since we
     know it is zero, we do not need to store it explicitely. */
  *lastp = 0;			/* init the last element of the array */
  scale = sqrt(1-pow(2.0,2.0*H-2.0)) * std; /* scaling factor for step 0 */
  for (i = 1; i <= n; i++)	/* generate the FBM in n steps */
    {
      d = 1 << (n-i);		/* distance between successive points */
      scale *= scalefactor;	/* scaling factor for step i */
      tracep[d-1] = 0.5 * tracep[2*d-1] + gasdev(random) * scale;
      for (j = 3*d-1; j < N-d; j += 2*d)
	tracep[j] = 0.5 * (tracep[j-d] + tracep[j+d]) + gasdev(random) * scale;
    }

  mean = M*T/N;			/* mean of the resulting trace */
  ampl = sqrt (a*M);		/* amplifier factor */
  for (sp = lastp; sp > tracep; sp--)
    {
      register float val = mean + ampl * (*sp - *(sp-1));
      *sp = max (val, 0);
    }
  *tracep = max (mean + ampl * (*tracep), 0);

  return (tracep);
}
