/****************************************************************************

  $Source: /home/foda/fracas/RCS/allocators.c $
  $Revision: 7.20 $
  $Date: 1998/07/24 10:02:17 $
  $State: Exp $
  $Locker:  $

****************************************************************************/
#ident "$Id: allocators.c 7.20 1998/07/24 10:02:17 pot Exp $"

/*
 *  FRACAS stands for FRAmed Channel Access Simulator.
 */

#include "fracas.h"

static allocator_r fixed_reader;
static allocator_r cfra_reader;
static allocator_r fodaibea_reader;
static allocator_r feeders_reader;
static allocator_r drifs_reader;

allocator_n allocators [] =
{
  { fixed_reader, "fixed" },
  { cfra_reader, "cfra" },
  { fodaibea_reader, "fodaibea" },
  { feeders_reader, "feeders" },
  { drifs_reader, "drifs" },
  { NULL }
};

/* tracing */
#define TRACE0(f) do if (trace && warmed) printf (f); while (0)
#define TRACE1(f,a) do if (trace && warmed) printf (f,a); while (0)
#define TRACE2(f,a,b) do if (trace && warmed) printf (f,a,b); while (0)
#define TRACE3(f,a,b,c) do if (trace && warmed) printf (f,a,b,c); while (0)
#define TRACE4(f,a,b,c,d) do if (trace && warmed) printf (f,a,b,c,d); while (0)
#define TRACE5(f,a,b,c,d,e) do if (trace && warmed) printf (f,a,b,c,d,e); while (0)

static logical trace;


/*
 * Fixed allocator.
 * Divides the space equally among the stations, independently of their
 * requests.
 *
 * Syntax is: allocator fixed
 *	The frame quote to divide among stations for datagram
 *	allocation is equal to the frame length minus the stream
 *	and vbrmin requests.
 */
static struct
{
  int sum_s_v;
  int fixed_d;
} fixed;

static void
fixed_allocator (void)
{
  station *stp;

  /* Initialise the sum of stream and vbr allocations.
     Do this only once. */
  if (fixed.sum_s_v < 0)
    {
      fixed.sum_s_v = 0;
      for_all_stations
	fixed.sum_s_v += stp->sreq + stp->vmaxreq;
      fixed.fixed_d = (framesize - fixed.sum_s_v) / stno;
    }

  for_all_stations
    {
      stp->all_f->allocation.s = stp->sreq;
      stp->all_f->allocation.v = stp->vmaxreq;
      stp->all_f->allocation.i = fixed.fixed_d;
      stp->all_f->allocation.b = 0;
    }
} /* fixed_allocator() */

static allocator *
fixed_reader (char *lp)
{
  memset ((char *)&fixed, 0, sizeof(fixed));
  fixed.sum_s_v = -1;

  if (cyclelen < 2*(excess_intdiv(rttime,frametime)+1) + 2)
    return_error ("cyclelen is too small for fixed allocator to work.");
  all_delay = 2*(excess_intdiv(rttime,frametime)+1);

  return (&fixed_allocator);
} /* fixed_reader() */

/*
 * CFRA allocator.
 * Uses the CFRA algorithm for the datagram transmissions.
 *
 * Syntax is: allocator cfra [slots] [Rmin] [Rmax] [trace]
 *	Where the optional <trace> argument is either `y' or `n' to enable
 *	or disable the feature (disabled by default).
 *
 *	This function performs allocation of datagram frame space between
 *	the stations which have made requests for it.  The function is
 *      called only after the stream allocation is done.  Stream and
 *	vbr traffics get the requested allocation.
 *
 *	<slots> is the number of data slots per frame, default 128.
 * 	<Rmin> and <Rmax> are the minimum and maximum number of slots
 * 	allocated to a station; defaults are 1 and 30.
 */
typedef struct vsat
{
  struct vsat *v_prev;		/* prev AR in the queue */
  struct vsat *v_next;		/* next AR in the queue */
  station *v_stp;		/* pointer to the vsat's station */
  logical v_AR;			/* Allocation Request is queued */
  logical v_ended;		/* `E' message received */
  int v_credit;
} vsat;

static struct
{
  unsigned int slots;		/* number of data slots in a frame */
  unsigned int sslots;		/* slots used for stream traffic */
  float slot_size;		/* size of a slot in TRUs */
  unsigned int Rmin, Rmax;	/* min and max allocation for a station */
  float min_alloc;		/* minimum allocation */
  vsat *vsat;			/* array of stno+1 vsat's (first unused) */
  vsat *first_AR;		/* first vsat in the AR queue */
  vsat *last_AR;		/* last vsat in the AR queue */
  int initial_credit;		/* first (negative) credit given to new ARs */
  int d_index;			/* index of first station tp get the gift */
} cfra;

/* Add the AR at the end of the queue. */
static void
queue_vsat (vsat *vsatp)
{
  ASSERT (!vsatp->v_AR);
  vsatp->v_next = NULL;
  if (cfra.first_AR == NULL)
    {
      vsatp->v_prev = NULL;
      cfra.first_AR = vsatp;
    }
  else
    {
      vsatp->v_prev = cfra.last_AR;
      cfra.last_AR->v_next = vsatp;
    }
  cfra.last_AR = vsatp;
  vsatp->v_AR = TRUE;
}

/* Remove the AR from the queue.  Don't reset the pointers inside the
   vsat, so the for loop may continue to scan the queue. */
static void
dequeue_vsat (vsat *vsatp)
{
  ASSERT (vsatp->v_AR);
  if (cfra.first_AR == vsatp)
    cfra.first_AR = vsatp->v_next;
  else
    vsatp->v_prev->v_next = vsatp->v_next;
  if (cfra.last_AR == vsatp)
    cfra.last_AR = vsatp->v_prev;
  else
    vsatp->v_next->v_prev = vsatp->v_prev;
  vsatp->v_AR = FALSE;
}

static void
cfra_allocator ()
{
  int j, free_slots;
  vsat *vsatp;


  /* First process all `S' and `E' messages. */
  TRACE1 ("\n*processing S and E messages at time %d.\n", curtime);
  for (j = 1; j <= stno; j++)
    {
      int dreq = st_table[j].curr_f->request.i;
      vsatp = &cfra.vsat[j];

      if (dreq == 0) /* `S' message */
	{
	  if (!vsatp->v_AR) /* AR not already queued */
	    {
	      TRACE1 ("*S msg from station %d (not queued)\n", j);
	      queue_vsat (vsatp);
	      vsatp->v_credit = cfra.initial_credit;
	    }
	  else
	    {
	      TRACE2 ("*S msg from station %d (credit was %d)\n",
		      j, vsatp->v_credit);
	    }
	  vsatp->v_ended = FALSE;
	}
      else if (dreq > 0) /* `E' message */
	{
	  ASSERT (vsatp->v_AR);
	  TRACE3 ("*E msg len=%d from station %d (credit was %d)\n",
		  dreq, j, vsatp->v_credit);
	  vsatp->v_credit += dreq;
	  vsatp->v_ended = TRUE;
	}
    }

  /* For datagram, give Rmin allocation to everyone, then
     scan the AR queue and give further allocation. */
  for (j = 1; j <= stno; j++)
    {
      station *stp = &st_table[j];
      itraffic *allp = &stp->all_f->allocation;

      allp->s = stp->sreq;
      allp->v = 0;
      allp->i = cfra.min_alloc;
      allp->b = 0;
    }
  TRACE2 ("*allocation %d given to %d stations\n", (int)cfra.min_alloc, stno);
  free_slots = cfra.slots - cfra.sslots - cfra.Rmin * stno;

  /* Scan the stations in the AR queue until all requests scanned
     or no more free slots. */
  for (vsatp = cfra.first_AR;
       vsatp != NULL && free_slots > 0;
       vsatp = vsatp->v_next)
    {
      int dalloc_more;

      vsatp->v_credit -= cfra.min_alloc;
      if (vsatp->v_ended)	/* burst already ended */
	{
	  if (vsatp->v_credit <= 0)
	    {
	      dequeue_vsat (vsatp);
	      dalloc_more = 0;
	    }
	  else
	    {
	      dalloc_more = MIN (vsatp->v_credit, cfra.Rmax-cfra.Rmin);
	    }
	}
      else			/* burst not yet ended */
	{
	  dalloc_more = cfra.Rmax-cfra.Rmin;
	}
      if (dalloc_more > 0)
	{
	  dalloc_more = MIN (dalloc_more, free_slots);
	  vsatp->v_credit -= floor (0.5 + dalloc_more * cfra.slot_size);
	  TRACE3 ("*further %d allocation to station %ld (credit now %d)\n",
		  (int) floor (0.5 + dalloc_more * cfra.slot_size),
		  vsatp->v_stp->number, vsatp->v_credit);
	  free_slots -= dalloc_more;
	}
      else
	TRACE1 ("*station %ld dequeued\n", vsatp->v_stp->number);

      /* Override the previous Rmin allocation. */
      vsatp->v_stp->all_f->allocation.i
	= floor (0.5 + ((cfra.Rmin + dalloc_more) * cfra.slot_size));
    }
  if (trace)
    {
      if (free_slots > 0)
	TRACE3
	  ("*distributing %d TRUs (%d free slots) starting from station %d\n",
	   (int)(free_slots*cfra.slot_size), free_slots, cfra.d_index);
      else
	{
	  TRACE0 ("*No free slots");
	  if (vsatp == NULL)
	    TRACE0 (" left\n");
	  else
	    {
	      TRACE0 ("; stations");
	      for (; vsatp != NULL; vsatp = vsatp->v_next)
		TRACE1 (" %ld", vsatp->v_stp->number);
	      TRACE0 (" unsatisfied\n");
	    }
	}
    }

  /* Distribute the remainder. */
  while (free_slots-- > 0)
    {
      st_table[cfra.d_index].all_f->allocation.i += cfra.slot_size;
      if (++cfra.d_index > stno)
	cfra.d_index = 1;
    }
} /* cfra_allocator() */


static allocator *
cfra_reader (char *lp)
{
  char *errstr;
  arg_type args[] =
  {
    { "slots", uns_t, optional },
    { "Rmin", uns_t, optional },
    { "Rmax", uns_t, optional },
    { "trace", log_t, optional },
    { NULL }
  };
  int j;

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("slots", cfra.slots);
  get_arg ("Rmin", cfra.Rmin);
  get_arg ("Rmax", cfra.Rmax);
  get_arg ("trace", trace);

  if (cfra.slots == 0)
    cfra.slots = 128;
  if (cfra.Rmin == 0)
    cfra.Rmin = 1;
  if (cfra.Rmax == 0)
    cfra.Rmax = 30;

  /* Initialisation. */
  cfra.slot_size = (float)framesize / cfra.slots;
  cfra.vsat = allocate ((stno+1) * sizeof (*cfra.vsat));
  cfra.first_AR = NULL;
  cfra.sslots = 0;
  for (j = 1; j <= stno; j++)
    {
      station *stp = &st_table[j];
      cfra.vsat[j].v_stp = stp;
      cfra.sslots += ceil (stp->sreq / cfra.slot_size);
    }
  cfra.d_index = 1;

  if (cfra.Rmin * stno > (cfra.slots - cfra.sslots))
    return_error ("Rmin*stno must not exceed the number of datagram slots");
  cfra.min_alloc = floor (0.5 + (cfra.Rmin * cfra.slot_size));
  if (cfra.Rmax < cfra.Rmin)
    return_error ("Rmax must not be less than Rmin");

  cfra.initial_credit = - cfra.Rmin * ((int)ceil(rttime/frametime) + 1);

  if (cyclelen < 2*(excess_intdiv(rttime,frametime)+1)+1 + 2)
    return_error ("cyclelen is too small for cfra allocator to work.");
  all_delay = 2*(excess_intdiv(rttime,frametime)+1)+1;

  return (&cfra_allocator);
} /* cfra_reader */

/*
 * FODA/IBEA allocator.
 * Uses the FODA/IBEA algorithm for the datagram transmissions.
 *
 * Syntax is: allocator fodaibea bovh cs csovh [fas] [<fas_every>]
 *				 [mindall] [maxdall] [csevery]
 *				 [allnmin] [allnmax] [alld]
 *				 [trace]
 *	Where the optional <trace> argument is either `y' or `n' to enable
 *	or disable the feature (disabled by default).
 *
 *	This function performs allocation of datagram frame space between
 *	the stations which have made requests for it.  The function is
 *      called only after the stream allocation is done.  Stream and
 *	vbr traffics get the requested allocation.
 *
 *	<cs> and <fas> are the sizes in TRUs of the control slot and first
 *	access slot respectively, <fas_every> is the number of frames from
 *	a fas to the next one.  <fas> defaults to 0, <fas_every> to 32.
 *	<bovh> is overhead needed by a station to transmit, in TRUs.
 *	<csovh> is overhead when only the control slot is allocated.
 *	<mindall> and <maxdall> are the min and max datagram allocations,
 *	respectively (in TRUs).  They default to bovh and framesize/2.
 *	The minimum and maximum quote of datagram request assigned are
 *	respectively (allnmin/alld) and (allnmax/alld).  The defaults
 *	are allnmin=5, allnmax=50, alld=100.
 *	One control slot (at least one) is assigned every <csevery>
 *	stations, defaulting to 8.
 */

/* miscellanea */
#define MIN_SPEED 1
#define FAKE_SPEED 1
#define LOW_SPEED MIN_SPEED
#define CS_SPEED MIN_SPEED	/* control slot speed */

#define MAX(a,b)	((a) > (b) ? (a) : (b))

#define FOR_ALL_SCB(scbp) \
  for ((scbp) = &g.scb[1]; (scbp) <= &g.scb[stno]; (scbp)++)

/* The channel request structure */
typedef struct
{
  int csbs;			/* CSB speed */
  int tru;			/* # of traffic units requested */
} request;

/* The station control block */
struct scb
{
  struct scb *s_next;		/* pointer to the next scb in the dring */
  struct scb *s_prev;		/* pointer to the previous scb in the dring */
  request s_dreq;		/* datagram request (TRUs) */
  int s_dreq_tru;		/* datagram request (TRUs) */
  request s_strass;		/* stream space assigned */
  request s_vbrass;		/* vbr space assigned */
  request s_dtgass;		/* datagram space assigned */
  int dovh_tru;			/* overhead used by datagram */
  int S_ACTIVE;			/* station is active */
  int S_DREQ;			/* station inserted in datagram ring */
};

static struct
{
  struct scb *scb;		/* station control block array */
  struct scb *cur_scbp;		/* ptr to the first scb in the datagram ring */
  int dring_count;		/* number of scb in the datagram ring */
  int unassigned_tru;		/* TRU to be assigned to first queued station*/
  struct scb *idle_scbp;	/* last st to get ctrl slot */
  int rrst;			/* first station to get the final gift */
  int bovh_tru;
  int cs_tru;
  int csovh_tru;
  int fas_every;
  int fas_tru;
  int min_dtga_tru;
  int max_dtga_tru;
  int allnmin, allnmax, alld;
  int csevery;
} g;

/*
 * Remove scb from the datagram ring
 *
 * input:  scb address
 * output: none
 */
static void
remove_scb_from_dring (struct scb *scbp)
{
  scbp->s_prev->s_next = scbp->s_next;
  scbp->s_next->s_prev = scbp->s_prev;
  scbp->S_DREQ = FALSE;

  if (--g.dring_count == 0)
    g.cur_scbp = NULL;			/* ring is now empty */
  else
    if (g.cur_scbp == scbp)
      g.cur_scbp = scbp->s_prev;	/* back to the previous */

  scbp->s_prev = NULL;
  scbp->s_next = NULL;
} /* remove_scb_from_dring() */


/*
 * Insert scb in the datagram ring, after the current position
 *
 * input:  scb address
 * output: none
 */
static void
insert_scb_in_dring_after (struct scb *scbp)
{
  if (g.dring_count > 0)
    /* ring not empty */
  {
    g.cur_scbp->s_next->s_prev = scbp;
    scbp->s_next = g.cur_scbp->s_next;

    g.cur_scbp->s_next = scbp;
    scbp->s_prev = g.cur_scbp;
  }

  else
    /* ring empty */
  {
    scbp->s_next = scbp;
    scbp->s_prev = scbp;
    g.cur_scbp = scbp;
  }

  scbp->S_DREQ = TRUE;
  g.dring_count++;

} /* insert_scb_in_dring_after() */


/*
 * Insert scb in the datagram ring, before the current position
 *
 * input:  scb address
 * output: none
 */
static void
insert_scb_in_dring_before (struct scb *scbp)
{
  if (g.dring_count > 0)
    /* ring not empty */
  {
    g.cur_scbp->s_prev->s_next = scbp;
    scbp->s_prev = g.cur_scbp->s_prev;

    g.cur_scbp->s_prev = scbp;
    scbp->s_next = g.cur_scbp;
  }

  else
    /* ring empty */
  {
    scbp->s_next = scbp;
    scbp->s_prev = scbp;
    g.cur_scbp = scbp;
  }

  scbp->S_DREQ = TRUE;
  g.dring_count++;

} /* insert_scb_in_dring_before() */


/*
 * fodaibea:
 *	 This function performs allocation of datagram frame space between
 *	 the stations which have made requests for it.  All these stations
 *	 are linked in the datagram ring.  The function is called only after
 *	 the stream allocation is done.
 */
static void
fodaibea_allocator ()
{
  logical fas_present = (curframeno % g.fas_every == 0);
  register struct scb *scbp;	/* station control block pointer */
  struct scb *first_idle_scbp;
  int st_count;			/* stations in dring before the loop begins */
  int cslots;			/* number of control slots to assign */
  int tot_str_tru;		/* total size of stream allocations */
  int avail_tru;		/* free TRUs in the frame */
  int gift_tru;			/* unrequested TRUs to give to each st */
  int ass_tru, ovh_tru;		/* temporary allocation and overhead values */
  int happy_st, angry_st;	/* # of satisfied and not satisfied st */
  int speed;			/* datagram csbs */
  int j, used_tru;


  /* Add to the datagram ring all the stations which have made a non null
     datagram request.  Compute the total size of stream allocations. */
  tot_str_tru = 0;
  for (j = 1; j <= stno; j++)
    {
      station *stp = &st_table[j];
      int str_ass = stp->sreq;
      int vbr_ass = stp->curr_f->request.v;

      scbp = &g.scb[j];

      if (vbr_ass < stp->vminreq)
	vbr_ass = stp->vminreq;
      else if (vbr_ass > stp->vmaxreq)
	vbr_ass = stp->vmaxreq;
      scbp->s_strass.csbs = FAKE_SPEED;
      scbp->s_strass.tru = str_ass;
      scbp->s_vbrass.tru = vbr_ass;
      tot_str_tru += str_ass + vbr_ass;
      if (str_ass > 0 || vbr_ass > 0)
	tot_str_tru += g.bovh_tru;
      if (stp->curr_f->request.i != NO_REQUEST)
	{
	  scbp->s_dreq.csbs = FAKE_SPEED;
	  scbp->s_dreq_tru = stp->curr_f->request.i;
	  if (!scbp->S_DREQ)	/* not in datagram ring */
	    {
	      if (stp->curr_f->request.i > 0)	/* and non null request */
		{
		  if (g.unassigned_tru == 0)	/* no rest to assign */
		    {
		      /* Insert the scb before the current one and make
			 it current. */
		      insert_scb_in_dring_before (scbp);
		      g.cur_scbp = scbp;
		    }
		  else
		    {
		      /* Insert the scb after the current one, because
			 the current has right to unassigned_tru. */
		      insert_scb_in_dring_after (scbp);
		    }
		}
	    }
	  else			/* in datagram ring */
	    {
	      if (stp->curr_f->request.i == 0)	/* null request */
		{
		  if (scbp == g.cur_scbp)
		    g.unassigned_tru = 0;
		  /* Warning: the following function call modifies cur_scbp
		     if it is equal to the function argument. */
		  remove_scb_from_dring (scbp);
		}
	    }
	}
    }



  /*
   * FODA datagram allocation algorithm
   */

  /* Compute the available frame space after the stream allocations. */
  cslots = (stno - 1) / g.csevery + 1;
  avail_tru = framesize - tot_str_tru - cslots * g.cs_tru;
  if (fas_present)
    {
      avail_tru -= g.fas_tru;
      do
	{
	  cslots -= 1;
	  avail_tru += g.cs_tru;
	}
      while (cslots > 0 && avail_tru <= 0);
    }
  ASSERT (avail_tru >= 0);

  TRACE3 ("\n*at time %d g.tot_str_tru = %d, avail_tru = %d.\n",
	  curtime, tot_str_tru, avail_tru);

  /* Reset allocations. */
  FOR_ALL_SCB (scbp)
    {
      scbp->s_dtgass.tru = 0;
      scbp->dovh_tru = 0;
    }

  /*
   * Assignment loop.
   * Scan the datagram ring once, starting from the current station
   * control block, until the available space is exhausted or all the
   * station requests have been examinated.
   */
  TRACE2 ("*%d stations are queued, unassigned_tru = %d.\n",
	  g.dring_count, g.unassigned_tru);

  ASSERT (g.dring_count > 0 || g.unassigned_tru == 0);

  st_count = g.dring_count;
  if (st_count < g.allnmin) st_count = g.allnmin;
  else if (st_count > g.allnmax) st_count = g.allnmax;

  scbp = g.cur_scbp;

  for (j = g.dring_count + ((g.unassigned_tru == 0) ? 0 : 1);
       j > 0 && avail_tru > 0;
       j--)
    {
      /* Get the preamble bit rate. */
      speed = scbp->s_dreq.csbs;

      /*
       * It is necessary to check if overhead for datagram data burst is
       * needed.  If this is the first time this station is examined in
       * the loop, overhead must be added in the case that station has no
       * stream space assigned or its stream premble bit rate is higher
       * than the one requested for datagram.  It is also necessary to
       * compare the assignment with the minimum allocation.
       */

      if (scbp->s_dtgass.tru == 0
	  && ((scbp->s_strass.tru == 0 && scbp->s_vbrass.tru == 0)
	      || scbp->s_strass.csbs > speed))
	{
	  ovh_tru = g.bovh_tru;	/* overhead needed */
	  /*
	   * Are there enough available TRUs to give the station the minimum
	   * data allocation plus overhead?
	   * If not, exit from the assignment loop.  First assignment of
	   * next frame restarts from this station, because we do not
	   * advance the scbp pointer.
	   */
	  if (ovh_tru + g.min_dtga_tru > avail_tru)
	    break;
	}
      else
	ovh_tru = 0;		/* no overhead needed */

      /* Set the correct speed for this station. */
      scbp->s_dtgass.csbs = speed;

      /* Reduced assignment. */
      if (g.unassigned_tru > 0)
	{
	  /*
	   * We store in g.unassigned_tru our missing TRUs.  At the end of the
	   * loop, the scbp pointer is not advanced if g.unassigned_tru is
	   * not null, so in the next frame we'll restart from this station.
	   */
	  ass_tru = g.unassigned_tru;
	  g.unassigned_tru = 0;
	}
      else
	{
	  ass_tru = MAX (scbp->s_dreq_tru, 0) * st_count;
	  ass_tru /= g.alld;
	}

      /* Increase the assignment if it is too small. */
      if (ass_tru < g.min_dtga_tru)
	ass_tru = g.min_dtga_tru;

      /* Check the upper bound. */
      if (ass_tru > g.max_dtga_tru)
	ass_tru = g.max_dtga_tru;

      g.unassigned_tru = ass_tru + ovh_tru - avail_tru;
      if (g.unassigned_tru > 0)
	{
	  ass_tru -= g.unassigned_tru; /* reduce the assignment */
	  ASSERT ((ovh_tru == 0) || (ass_tru >= g.min_dtga_tru));
	}
      else
	g.unassigned_tru = 0;	/* rest is null */

      TRACE3
	("*station %d requests %d data traffic units, gets %d in the loop.\n",
	 scbp - g.scb, scbp->s_dreq_tru, ass_tru);

      /* Decrement station request and set allocation in scb. */
      scbp->s_dreq_tru -= ass_tru;
      scbp->s_dtgass.tru += ass_tru;
      scbp->dovh_tru = ovh_tru;

      /* Remaining available datagram space. */
      avail_tru -= ovh_tru + ass_tru;
      ASSERT (avail_tru >= 0);

      /* Check if the current station request is fully satisfied. */
      if (scbp->s_dreq_tru <= 0)
	{
	  struct scb *next_scbp = scbp->s_next;

	  /* Warning: the following function call modifies cur_scbp if
	     it is equal to the function argument. */
	  remove_scb_from_dring (scbp);
	  ASSERT ((g.cur_scbp == NULL) == (g.dring_count == 0));

	  if (g.dring_count == 0)
	    {
	      /* If the current station is the last in the datagram
		 ring and it has been completely satisfied, perhaps it
		 makes no sense to leave unassigned_tru set.  Anyway,
		 this is the simplest thing to do, because otherwise
		 one would have to leave the station in the ring when
		 unassigned_tru is not null. */
	      g.unassigned_tru = 0;
	      break;		/* exit from the assignment loop  */
	    }
	  else
	    scbp = next_scbp;	/* advance scb */

	  /* The station is completely satisfied,
	     so null the missing TRU count. */
	  g.unassigned_tru = 0;
	} /* station fully satisfied */
      else
	{
	  /* Advance the scbp pointer. */
	  ASSERT ((g.unassigned_tru == 0) || (avail_tru == 0));
	  if (g.unassigned_tru == 0)
	    scbp = scbp->s_next;
	}
    } /* end of assignment loop */


  TRACE2
    ("*End of assignment loop: %d stations queued, %d tru to be assigned.\n",
     g.dring_count, avail_tru);

  /* Set the starting point for the loop of the next frame. */
  if (g.dring_count > 0)
    g.cur_scbp = scbp;

  /* Assign the control slots. */
  scbp = first_idle_scbp = g.idle_scbp;
  while (cslots > 0)
    {
      /* Advance the station pointer. */
      if (++scbp == &g.scb[stno+1])
 	scbp = &g.scb[1];

      /* If the station has no assignemnts it gets a control slot.
	 The idle pointer is left pointing to that station. */
      if (scbp->s_strass.tru == 0
	  && scbp->s_vbrass.tru == 0
	  && scbp->s_dtgass.tru == 0)
 	{
	  TRACE2 ("*Station %d gets a control slot (%d traffic units).\n",
		  scbp - g.scb, g.cs_tru);
	  scbp->s_dtgass.csbs = CS_SPEED;
	  scbp->s_dtgass.tru = g.cs_tru - g.csovh_tru;
	  scbp->dovh_tru = g.csovh_tru;
	  scbp->s_dreq_tru = MAX (scbp->s_dreq_tru - g.cs_tru, 0);
	  g.idle_scbp = scbp;
	  cslots -= 1;
	  continue;
 	}

      /* If we have looked at all the stations, exit
 	 and distribute the unassigned control slots. */
      if (scbp == first_idle_scbp)
 	{
	  avail_tru += cslots * g.cs_tru;
	  break;
	}
    } /* assign the control slots */

  TRACE1 ("*Still %d traffic units to be assigned.\n", avail_tru);


  /*
   * If all the stations have been fully satisfied, some traffic units are
   * probably available and not distributed. In this case, a gift is
   * made to one of these classes of stations (in the given order),
   * only if the minimum allocation can be reached:
   *  if no angry station--all the stations that are in the datagram queue,
   *  all the stations,
   *  all the stations that did not receive any allocation
   *  all the stations that did already receive an allocation.
   */
  happy_st = 0;
  angry_st = 0;
  if (avail_tru > 0)
    {
      /* Compute how many stations did receive an allocation and how
	 many didn't receive any. */
      FOR_ALL_SCB (scbp)
	{
	  if (scbp->s_dtgass.tru > 0)
	    happy_st++;
	  else
	    angry_st++;
	}
      TRACE2 ("*%d angry and %d happy stations.\n", angry_st, happy_st);

      if (angry_st == 0 && g.dring_count > 0)
	{
	  gift_tru = avail_tru / g.dring_count;

	  /* Assign the quotient. */
	  FOR_ALL_SCB (scbp)
	    if (scbp->S_DREQ)
	      {
		scbp->s_dtgass.tru += gift_tru;
		avail_tru -= gift_tru;
		if ((scbp->s_dreq_tru -= gift_tru) <= 0)
		  {
		    if (scbp == g.cur_scbp)
		      g.unassigned_tru = 0;
		    remove_scb_from_dring (scbp);
		    ASSERT ((g.cur_scbp == NULL) == (g.dring_count == 0));
		  }
		TRACE2
	  ("*Station %d is in dgram queue, so gets %d TRU of data as gift.\n",
	   scbp - g.scb, gift_tru);
	      }
	}
      else if ((happy_st + angry_st > 0)
	       && ((gift_tru = ((avail_tru - g.bovh_tru * angry_st)
				/ (happy_st + angry_st)))
		   > g.min_dtga_tru))
	{
	  /* Assign the quotient. */
	  FOR_ALL_SCB (scbp)
	    {
	      if (scbp->s_dtgass.tru == 0)
		{
		  scbp->s_dtgass.csbs = LOW_SPEED;
		  scbp->dovh_tru = g.bovh_tru;
		  avail_tru -= g.bovh_tru;
		}
	      scbp->s_dtgass.tru += gift_tru;
	      avail_tru -= gift_tru;
	      if (scbp->S_DREQ)
		if ((scbp->s_dreq_tru -= gift_tru) <= 0)
		  {
		    if (scbp == g.cur_scbp)
		      g.unassigned_tru = 0;
		    remove_scb_from_dring (scbp);
		    ASSERT
		      ((g.cur_scbp == NULL) == (g.dring_count == 0));
		  }
	      TRACE2 ("*Station %d gets %d TRU of data as gift.\n",
		      scbp - g.scb, gift_tru);
	    }
	}
      else if (angry_st > 0
	       && ((gift_tru = avail_tru / angry_st - g.bovh_tru)
		   >= g.min_dtga_tru))
	{
	  /* Assign the quotient. */
	  FOR_ALL_SCB (scbp)
	    if (scbp->s_dtgass.tru == 0)
	      {
		scbp->s_dtgass.csbs = LOW_SPEED;
		scbp->s_dtgass.tru = gift_tru;
		scbp->dovh_tru = g.bovh_tru;
		avail_tru -= g.bovh_tru + gift_tru;
		if (scbp->S_DREQ)
		  if ((scbp->s_dreq_tru -= gift_tru) <= 0)
		    {
		      if (scbp == g.cur_scbp)
			g.unassigned_tru = 0;
		      remove_scb_from_dring (scbp);
		      ASSERT
			((g.cur_scbp == NULL) == (g.dring_count == 0));
		    }
		TRACE2
		  ("*Station %d is angry, so gets %d TRU of data as gift.\n",
		   scbp - g.scb, gift_tru);
	      }
	}

      else if (happy_st > 0
	       && (gift_tru = avail_tru / happy_st) > 0)
	{
	  /* Assign the quotient. */
	  FOR_ALL_SCB (scbp)
	    if (scbp->s_dtgass.tru > 0)
	      {
		scbp->s_dtgass.tru += gift_tru;
		avail_tru -= gift_tru;
		if (scbp->S_DREQ)
		  if ((scbp->s_dreq_tru -= gift_tru) <= 0)
		    {
		      if (scbp == g.cur_scbp)
			g.unassigned_tru = 0;
		      remove_scb_from_dring (scbp);
		      ASSERT
			((g.cur_scbp == NULL) == (g.dring_count == 0));
		    }
		TRACE2
		  ("*Station %d is happy, so gets %d TRU of data as gift.\n",
		   scbp - g.scb, gift_tru);
	      }
	}

  } /* make gift */

  /* Distribute the remainder. */
  if (happy_st > 0)
    while (avail_tru > 0)
      {
 	scbp = &g.scb[g.rrst];
 	if (scbp->s_dtgass.tru > 0)
	  {
	    scbp->s_dtgass.tru += 1;
	    avail_tru -= 1;
	  }
 	if (++g.rrst > stno)
	  g.rrst = 1;
      }

  /* Check the total number of assigned TRUs. */
  used_tru = 0;
  FOR_ALL_SCB (scbp)
    used_tru += scbp->s_strass.tru + scbp->s_dtgass.tru + scbp->dovh_tru;
  ASSERT (used_tru + (fas_present ? g.fas_tru : 0) <= framesize);

  /* Write the allocations in the proper places. */
  for (j = 1; j <= stno; j++)
    {
      st_table[j].all_f->allocation.s = g.scb[j].s_strass.tru;
      st_table[j].all_f->allocation.v = g.scb[j].s_vbrass.tru;
      st_table[j].all_f->allocation.i = g.scb[j].s_dtgass.tru;
      st_table[j].all_f->allocation.b = 0;

      ASSERT (st_table[j].all_f->allocation.i >= 0); }

} /* fodaibea_allocator() */


static allocator *
fodaibea_reader (char *lp)
{
  station *stp;
  char *errstr;
  int tot_str_tru;
  arg_type args[] =
  {
    { "bovh", uns_t, required },
    { "cs", uns_t, required },
    { "csovh", uns_t, required },
    { "fas", uns_t, optional },
    { "fas_every", uns_t, optional },
    { "mindall", uns_t, optional },
    { "maxdall", uns_t, optional },
    { "csevery", int_t, optional },
    { "allnmin", int_t, optional },
    { "allnmax", int_t, optional },
    { "alld", int_t, optional },
    { "trace", log_t, optional },
    { NULL }
  };

  /* Clean my local variables. */
  memset ((char *)&g, 0, sizeof(g));
  g.rrst = 1;

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("bovh", g.bovh_tru);
  get_arg ("cs", g.cs_tru);
  get_arg ("csovh", g.csovh_tru);
  get_arg ("trace", trace);
  get_arg ("fas", g.fas_tru);
  get_arg ("fas_every", g.fas_every);
  get_arg ("mindall", g.min_dtga_tru);
  get_arg ("maxdall", g.max_dtga_tru);
  get_arg ("allnmin", g.allnmin);
  get_arg ("allnmax", g.allnmax);
  get_arg ("alld", g.alld);
  get_arg ("csevery", g.csevery);

  if (g.bovh_tru >= framesize)
    return_error ("bovh must be less than framesize");
  if (g.cs_tru > framesize)
    return_error ("cs must be less than framesize");
  if (g.cs_tru < g.csovh_tru)
    return_error ("cs must be not less than csovh");
  if (g.fas_tru > framesize)
    return_error ("fas must be less than framesize");

  if (g.fas_every == 0) g.fas_every = 32;

  if (g.max_dtga_tru == 0)
    g.max_dtga_tru = framesize / 2;
  if (g.max_dtga_tru <= g.bovh_tru)
    return_error ("maxdall must be greater than bovh");

  if (g.min_dtga_tru == 0)
    g.min_dtga_tru = g.bovh_tru;
  if (g.min_dtga_tru > g.max_dtga_tru)
    return_error ("mindall must not be greater than maxdall");

  if (cyclelen < 2*(excess_intdiv(rttime,frametime)+1)+1 + 2)
    return_error ("cyclelen is too small for fodaibea allocator to work.");
  all_delay = 2*(excess_intdiv(rttime,frametime)+1)+1;

  if (g.allnmin == 0) g.allnmin = 5;
  if (g.allnmax == 0) g.allnmax = 50;
  if (g.alld == 0) g.alld = 100;
  if (g.csevery == 0) g.csevery = 8;
  if (g.allnmin < 0)
    return_error ("allnmin must not be negative");
  if (g.allnmax < g.allnmin)
    return_error ("allnmax must not be less than allnmin");
  if (g.alld <= 0)
    return_error ("alld must be positive");
  if (g.csevery <= 0)
    return_error ("csevery must be positive");

  tot_str_tru = 0;
  for_all_stations
    {
      int stream_all = stp->sreq + stp->vmaxreq;
      if (stream_all > 0)
	stream_all += g.bovh_tru;
      tot_str_tru += stream_all;
    }
  if (tot_str_tru + g.fas_tru > framesize)
    return_error
      ("stream and vbrmax requests plus overheads do not leave space for fas");
  if (tot_str_tru + g.cs_tru > framesize)
    return_error
      ("stream and vbrmax requests plus overheads do not leave space for cs");

  /* Initialisation. */
  g.scb = allocate ((stno+1) * sizeof (struct scb));
  g.idle_scbp = &g.scb[1];

  return (&fodaibea_allocator);
} /* fodaibea_reader */

/*
 * FEEDERS allocator.
 * Uses the FEEDERS algorithm for the datagram transmissions.
 *
 * Syntax is: allocator feeders dquote rbovh [bovh] [maxdall] [na] [trace]
 *	Where the optional <trace> argument is either `y' or `n' to
 *	enable or disable the feature.
 *
 *	This function performs allocation of datagram frame space between
 *	the stations which have made requests for it.  The function is
 *      called only after the stream allocation is done.  Stream and
 *	vbr traffics get the requested allocation.
 *
 *	<dquote> (a positive floating point number) is the quote of
 * 	the datagram request that is assigned in each loop, <rbovh> is
 * 	the length of the reference burst and <bovh> is the length of
 * 	a data burst overhead, both in TRUs, <maxdall> is the maximum
 *	assignement given before shrinking or enlargement of allocations,
 *	<na> is the number of frames in an allocation cycle.
 */
static struct feeders_v
{
  unsigned int fv_na;
  unsigned int fv_rbovh;
  unsigned int fv_bovh;
  unsigned int fv_maxdall;
  float fv_dquote;
  int *fv_dreq;
  int *fv_vallocation;
  int fv_maxdreq;		/* upper limit for the datagram request */
  int fv_ac_seq;		/* allocation cycle sequence no.[0..fv_na-1] */
  int fv_rrst;			/* first station to get the last gift */
} fv;

static void
feeders_allocator ()
{
  static int rest_tru;		/* part of frame to be divided evenly */
  static float dquote;		/* quote of the dgram request assigned */
  int avail_tru = 0;		/* free TRUs in the frame */
  int j;


  if (fv.fv_ac_seq == 0)
    /* this is the first frame in the allocation cycle */
    {
      int tot_dreq_q = 0;	/* total size of scaled datagram requests */
      int tot_dreq = 0;		/* total size of datagram requests */
      int tot_str_tru = 0;	/* total size of stream allocations */
      int tot_ovh = 0;		/* total burst overheads */

      /* Compute the total size of stream allocations and datagram requests. */
      avail_tru = 0;
      for (j = 1; j <= stno; j++)
        {
          station *stp = &st_table[j];
          int str_ass = stp->sreq;
          int vbr_ass = stp->curr_f->request.v;
          int dreq = MIN (st_table[j].curr_f->request.i, fv.fv_maxdreq);

          if (vbr_ass < stp->vminreq)
	    vbr_ass = stp->vminreq;
          else if (vbr_ass > stp->vmaxreq)
	    vbr_ass = stp->vmaxreq;

	  ASSERT (dreq != NO_REQUEST);

          if (str_ass > 0 || vbr_ass > 0)
	    tot_str_tru += str_ass + vbr_ass;
	  else
	    tot_ovh += fv.fv_bovh;
	  tot_dreq += dreq;
	  tot_dreq_q += dreq * fv.fv_dquote;
          fv.fv_vallocation[j] = vbr_ass;
	}

      /* Compute the available frame space in the first frame of the
	 allocation cycle after the stream allocations. */
      avail_tru = framesize - tot_str_tru - tot_ovh - fv.fv_rbovh;

      TRACE4
	("\n*at time %d: tot_str %d, tot_dreq %d, avail(first frame) %d\n",
	 curtime, tot_str_tru, tot_dreq, avail_tru);

      dquote = fv.fv_dquote;
      if (avail_tru < tot_dreq_q)
	{
	  TRACE1 ("*Shrinking by %g times\n", tot_dreq_q/(double)avail_tru);
	  dquote = avail_tru / (float)tot_dreq;
	}
      for (j = 1; j <= stno; j++)
	{
          int dreq_q = (MIN (st_table[j].curr_f->request.i, fv.fv_maxdreq)
			* dquote);

	  fv.fv_dreq[j] = dreq_q;
	  avail_tru -= dreq_q;
	}
      rest_tru = avail_tru;
      ASSERT (rest_tru >= 0);
    } /* if this is the first frame in the allocation cycle */

  TRACE3 ("*Rest to distribute in frame %d/%d of the all. cycle is %d\n",
	  fv.fv_ac_seq+1, fv.fv_na, rest_tru);

  /* Write the allocations in the proper places. */
  for (j = 1; j <= stno; j++)
    {
      station *stp = &st_table[j];
      stp->all_f->allocation.s = stp->sreq;
      stp->all_f->allocation.v = fv.fv_vallocation[j];
      stp->all_f->allocation.i = fv.fv_dreq[j] +  rest_tru / stno;
      stp->all_f->allocation.b = 0;
   }

  /* Distribute the remainder. */
  avail_tru = rest_tru % stno;
  if (avail_tru > 0)
    TRACE0 ("*1 TRU given to stations:");
  while (avail_tru > 0)
    {
      st_table[fv.fv_rrst].all_f->allocation.i += 1;
      avail_tru -= 1;
      TRACE1 (" %d", fv.fv_rrst);
      if (++fv.fv_rrst > stno)
	fv.fv_rrst = 1;
    }

  /* Total allocation. */
    {
      int tot_all = 0;
      TRACE0 ("\n");
      for (j = 1; j <= stno; j++)
	{
          TRACE2 ("*station %d gets %d\n", j, st_table[j].all_f->allocation.i);
	  tot_all += st_table[j].all_f->allocation.i;
	}
      TRACE1 ("*tot_all = %d\n", tot_all);
    }

  if (fv.fv_ac_seq == 0)
    rest_tru += fv.fv_rbovh;
 if (++fv.fv_ac_seq == fv.fv_na)
   fv.fv_ac_seq = 0;
} /* feeders_allocator() */

static allocator *
feeders_reader (char *lp)
{
  station *stp;
  int tot_ovh, tot_str_tru;
  char *errstr;
  arg_type args[] =
  {
    { "dquote", float_t, required },
    { "bovh", uns_t, optional },
    { "na", uns_t, required },
    { "rbovh", uns_t, required },
    { "maxdall", uns_t, optional },
    { "trace", log_t, optional },
    { NULL }
  };

  /* Clean my variables. */
  memset ((char *)&fv, 0, sizeof (fv));
  fv.fv_rrst = 1;

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("dquote", fv.fv_dquote);
  get_arg ("bovh", fv.fv_bovh);
  get_arg ("na", fv.fv_na);
  get_arg ("rbovh", fv.fv_rbovh);
  get_arg ("maxdall", fv.fv_maxdall);
  get_arg ("trace", trace);

  if (fv.fv_na == 0)
    return_error ("na must be positive");

  if (fv.fv_bovh >= framesize)
    return_error ("bovh must be less than framesize");

  if (fv.fv_rbovh == 0)
    return_error ("rbovh must be positive");
  if (fv.fv_rbovh >= framesize)
    return_error ("rbovh must be less than framesize");

  if (fv.fv_dquote <= 0)
    return_error ("dquote must be positive");

  fv.fv_dreq = allocate ((stno+1) * (sizeof(*fv.fv_dreq)));
  fv.fv_vallocation = allocate ((stno+1) * (sizeof(*fv.fv_vallocation)));

  if (cyclelen < excess_intdiv(rttime,frametime)+1+1 + 2)
    return_error ("cyclelen is too small for feeders allocator to work.");
  all_delay = excess_intdiv(rttime,frametime)+1+1;

  if (fv.fv_maxdall == 0)
    fv.fv_maxdall = (framesize - stno * fv.fv_bovh) / 2;
  fv.fv_maxdreq = fv.fv_maxdall / fv.fv_dquote;

  tot_str_tru = 0;
  tot_ovh = fv.fv_rbovh;
  for_all_stations
    {
      int str_req = stp->sreq + stp->vmaxreq;
      if (str_req == 0)
	tot_ovh += fv.fv_bovh;
      else
	tot_str_tru += str_req;
    }

  if (tot_str_tru + tot_ovh > framesize)
    return_error
    ("stream and vbrmax requests must be less than framesize minus overheads");

  return (&feeders_allocator);
} /* feeders_reader */

/*
 * Syntax is: allocator drifs dquote stinframe stovh [bovh] [maxdall] [trace]
 *	Where the optional <trace> argument is either `y' or `n' to
 *	enable or disable the feature.
 *
 *	This function performs allocation of datagram frame space between
 *	the stations which have made requests for it.  The function is
 *      called only after the stream allocation is done.  Stream and
 *	vbr traffics get the requested allocation.
 *
 *	<dquote> (a positive floating point number) is the quote of
 * 	the datagram request that is assigned in each loop, <stovh> is
 * 	the length of the control burst and <bovh> is the length of
 * 	a data burst overhead, both in TRUs, <stinframe> is the max number
 *  	of control bursts in a frame, <maxdall> is the maximum assignement
 *	given before shrinking or enlargement of allocations.
 */
static struct drifs_v
{
  unsigned int dv_stinframe;
  unsigned int dv_stovh;
  unsigned int dv_bovh;
  unsigned int dv_maxdall;
  float dv_dquote;
  int *dv_dallocation;
  int *dv_vallocation;
  int dv_na;			/* allocation cycle length in frames */
  int dv_maxdreq;		/* upper limit for the datagram request */
  int dv_ac_seq;		/* allocator cycle sequence no. [0..dv_na-1] */
  int dv_rrst;			/* first station to get the last gift */
} dv;

static void
drifs_allocator ()
{
  static int tot_str_tru;	/* total size of stream allocations */
  static int dvstno;		/* current station number */
  int avail;			/* rest of the available space in the frame */
  int j, k;


  /*
   * In the first frame of each allocation cycle, compute the allocations
   * for the whole allocation cycle.
   */
  if (dv.dv_ac_seq == 0)
    {
      int ac_tot_bovh = 0;	/* total size of burst overhead */
      int ac_tot_dreq = 0;	/* total size of datagram requests */
      int ac_tot_dreq_q = 0;	/* total size of requested allocations */
      int ac_avail = 0;		/* free TRUs in the allocation frame */
      int rest;

      tot_str_tru = 0;
      dvstno = 1;


      /* Compute the total size of stream allocations and datagram requests. */
      for (j = 1; j <= stno; j++)
        {
	  station *stp = &st_table[j];
          int str_ass = stp->sreq;
          int vbr_ass = stp->curr_f->request.v;
	  int dreq = MIN (stp->curr_f->request.i, dv.dv_maxdreq);
	  int dreq_q = dreq * dv.dv_na * dv.dv_dquote;


          if (vbr_ass < stp->vminreq)
	    vbr_ass = stp->vminreq;
          else if (vbr_ass > stp->vmaxreq)
	    vbr_ass = stp->vmaxreq;

	  ASSERT (dreq != NO_REQUEST);

          if (str_ass > 0 || vbr_ass > 0)
	    tot_str_tru += str_ass + vbr_ass;
	  else if (dreq_q > 0)
	    ac_tot_bovh += dv.dv_bovh;
	  if (dreq_q > 0)
	    {
	      ac_tot_dreq += dreq;
	      ac_tot_dreq_q += dreq_q;
	    }

	  dv.dv_vallocation[j] = vbr_ass;
	}

      /* Compute the frame space available after the stream allocations. */
      ac_avail = ((framesize - tot_str_tru) * dv.dv_na
		  - stno * dv.dv_stovh
		  - (dv.dv_na - 1) * dv.dv_bovh
		  - ac_tot_bovh);

TRACE5("\n*at time %d: tot_str %d, tot_dreq_q %d, ac_avail %d, tot_bovh %d.\n",
	curtime, tot_str_tru, ac_tot_dreq_q, ac_avail, ac_tot_bovh);

      /* Divide the rest evenly among the stations
	 or else shrink the requests. */
      if ((rest = ac_avail - ac_tot_dreq_q) > 0)
	{
	  /* Make a gift in the given order:
	   *  all the stations,
	   *  all the stations that did not receive any datagram allocation
	   *  all the stations that did already receive a dtgram allocation.
	   */
	  int happy_st = 0;	/* stations that did receive a dgram all. */
	  int unhappy_st = 0;	/* stations that did not receive a dgram all.*/
	  int angry_st = 0;	/* stations that did not receive any all. */
	  int gift_tru = 0;

	  /* Compute how many stations did receive a datagram allocation
	     and how many didn't receive any, then give the allocation. */
	  for (j = 1; j <= stno; j++)
	    {
	      int dreq_q = (MIN (st_table[j].curr_f->request.i, dv.dv_maxdreq)
			    * dv.dv_na * dv.dv_dquote);
	      if (dreq_q > 0)
		happy_st += 1;
	      else if (dv.dv_vallocation[j] > 0 || st_table[j].sreq > 0)
		unhappy_st += 1;
	      else
		angry_st += 1;
	      dv.dv_dallocation[j] = dreq_q;
	    }
	  TRACE3 ("*%d happy, %d unhappy, %d angry stations.\n",
		  happy_st, unhappy_st, angry_st);

	  /* Try all the stations first. */
	  if ((gift_tru = (rest - angry_st * dv.dv_bovh) / stno) > 0)
	    {
	      TRACE1 ("*gift of %d TRUs to all stations\n", gift_tru);
	      for (j = 1; j <= stno; j++)
		dv.dv_dallocation[j] += gift_tru;
	      ac_avail -= angry_st * dv.dv_bovh;
	    }

	  /* Try the stations that did not receive a datagram allocation. */
	  else if (unhappy_st + angry_st > 0
		   && ((gift_tru = (rest - angry_st * dv.dv_bovh)
			/ (unhappy_st + angry_st))
		       > 0))
	    {
	      TRACE1 ("*gift of %d TRUs to unhappy and angry stations\n",
		      gift_tru);
	      for (j = 1; j <= stno; j++)
		if (dv.dv_dallocation[j] == 0)
		  dv.dv_dallocation[j] = gift_tru;
	      ac_avail -= angry_st * dv.dv_bovh;
	    }

	  /* As last resort, try those that already have a dtg allocation. */
	  else if (happy_st > 0 && (gift_tru = rest / happy_st) > 0)
	    {
	      TRACE1 ("*gift of %d TRUs to happy stations\n", gift_tru);
	      for (j = 1; j <= stno; j++)
		if (dv.dv_dallocation[j] > 0)
		  dv.dv_dallocation[j] += gift_tru;
	    }
	} /* distribute the unrequested space */

      else
	/* not enough space to satisfy the requests: shrink them */
	{
	  /* While shrinking, rounding errors may occur.  So we must
	     recompute ac_avail. */
	  int sum = ac_avail;
	  float dquote = ac_avail / (float)ac_tot_dreq;

	  TRACE1 ("*Shrinking the requests by %g times\n",
		  (dv.dv_na * dv.dv_dquote) / dquote);

	  /* Put the overhead back into ac_avail, then subtract it
	     for each station that has a non null allocation. */
	  ac_avail += ac_tot_bovh;

	  for (j = 1; j <= stno; j++)
	    {
	      int dreq_q = (MIN (st_table[j].curr_f->request.i, dv.dv_maxdreq)
			    * dquote);

	      dv.dv_dallocation[j] = dreq_q;
	      if (dreq_q > 0)
		{
		  ac_avail -= dv.dv_bovh;
		  sum -= dreq_q;
		}
	    }
	  ASSERT (sum >= 0);
	  TRACE1 ("*Recomputed ac_avail after shrinking is %d\n", ac_avail);
	} /* shrink the requests */


      /* Distribute the remainder. */
      {
        int remaind = ac_avail;
	int happy_st = 0;

	for (j = 1; j <= stno; j++)
	  {
	    int dall = dv.dv_dallocation[j];
	    if (dall > 0)
	      {
		remaind -= dall;
		happy_st += 1;
	      }
	  }
	TRACE1 ("*remainder to distribute = %d\n", remaind);
	ASSERT (remaind >= 0);

	if (remaind > 0)
	  TRACE0 ("*1 TRU given to stations:");
	if (happy_st > 0)
	  while (remaind > 0)
	    {
	      if (dv.dv_dallocation[dv.dv_rrst] > 0)
		{
		  dv.dv_dallocation[dv.dv_rrst] += 1;
		  remaind -= 1;
		  TRACE1 (" %d", dv.dv_rrst);
		}
	      if (++dv.dv_rrst > stno)
		dv.dv_rrst = 1;
	    }

	/* Tracing printouts. */
	TRACE0 ("\n*Computed allocations for the allocation cycle:\n");
	for (j = 1; j <= stno && dv.dv_dallocation[j] == 0; j++)
	  continue;
	if (j == 2) j = 1;
	for (k = stno; j > 0 && dv.dv_dallocation[k] == 0; k--)
	  continue;
	if (k == stno-1) k = stno;
	if (j > 1)
	  TRACE1 ("* stations 1:%d get 0\n", j-1);
	for (; j <= k; j++)
	  TRACE2 ("* station %d gets %d\n",
		  j, dv.dv_dallocation[j]);
	if (k < stno)
	  TRACE2 ("* stations %d:%d get 0\n", k+1, stno);
      } /* distribute the remainder */
    } /* if this is the first frame in the allocation cycle */


  /*
   * Now compute the allocations for this frame.
   */
  for (j = 1; j <= stno; j++)
    {
      station *stp = &st_table[j];
      stp->all_f->allocation.s = stp->sreq;
      stp->all_f->allocation.v = dv.dv_vallocation[j];
      stp->all_f->allocation.i = 0;
      stp->all_f->allocation.b = 0;
    }

  avail = framesize - tot_str_tru - dv.dv_stinframe * dv.dv_stovh;
  if (dv.dv_ac_seq == dv.dv_na - 1)
    avail += (dv.dv_na * dv.dv_stinframe - stno) * dv.dv_stovh;

  TRACE2 ("\n*at time %d: avail %d\n", curtime, avail);

  for (; avail > 0 && dvstno <= stno; dvstno++)
    {
      station *stp = &st_table[dvstno];

      /* Skip this station if it has no datagram allocation. */
      if (dv.dv_dallocation[dvstno] == 0)
	continue;

      /* If the station has no stream or vbr allocations, then keep the
	 bovh into account. */
      if (stp->sreq == 0 && stp->vminreq == 0)
	{
	  if (avail > dv.dv_bovh)
	    {
	      avail -= dv.dv_bovh;
	    }
	  else		/* not enough space left for bovh */
	    {
	      /* Look for the previous station that has a datagram
		 allocation and give the available space to it.  Then
		 give the space unused for the overhead to the current
		 station. */
	      for (k = dvstno-1; k > 0; k--)
		if (dv.dv_dallocation[k] > 0)
		  break;
	      ASSERT (k > 0);

	      dv.dv_dallocation[k] += avail;
	      dv.dv_dallocation[dvstno] += dv.dv_bovh - avail;
	      TRACE4 ("*Station %d gains %d, station %d gains %d\n",
		      k, avail, dvstno, dv.dv_bovh-avail);

	      /* Remake the assignement to station k. */
	      avail = 0;
	      dvstno = k;
	    }
	}

      if (dv.dv_dallocation[dvstno] <= avail)
	{
	  avail -= stp->all_f->allocation.i = dv.dv_dallocation[dvstno];
	  if (dv.dv_ac_seq < dv.dv_na-1 && avail == 0)
	    {
	      /* Give to the next station the space unused for the overhead. */
	      for (k = dvstno+1; k <= stno; k++)
		if (dv.dv_dallocation[k] > 0)
		  break;
	      ASSERT (k <= stno);

	      dv.dv_dallocation[k] += dv.dv_bovh;
	      TRACE2 ("*Station %d gains %d\n", k, dv.dv_bovh);
	    }
	}
      else
        {
          dv.dv_dallocation[dvstno] -= avail;
          stp->all_f->allocation.i = avail;
	  avail = 0;
	  break;		/* don't advance dvstno, exit the for loop */
	}
    } /* for all the stations that should have an allocation in this frame */

  /* Tracing printouts. */
  TRACE2 ("*Real allocations for frame %d/%d of allocation cycle:\n",
	  dv.dv_ac_seq+1, dv.dv_na);
  for (j = 1; j <= stno && st_table[j].all_f->allocation.i == 0; j++)
    continue;
  if (j == 2) j = 1;
  for (k = stno; j > 0 && st_table[k].all_f->allocation.i == 0; k--)
    continue;
  if (k == stno-1) k = stno;
  if (j > 1)
    TRACE1 ("* stations 1:%d get 0\n", j-1);
  for (; j <= k; j++)
    TRACE2 ("* station %d gets %d\n", j, st_table[j].all_f->allocation.i);
  if (k < stno)
    TRACE2 ("* stations %d:%d get 0\n", k+1, stno);

  ASSERT (avail == 0);

 if (++dv.dv_ac_seq == dv.dv_na)
   dv.dv_ac_seq = 0;
} /* drifs_allocator() */

static allocator *
drifs_reader (char *lp)
{
  station *stp;
  char *errstr;
  int tot_str_tru;
  arg_type args[] =
  {
    { "dquote", float_t, required },
    { "bovh", uns_t, optional },
    { "stinframe", uns_t, required },
    { "stovh", uns_t, required },
    { "maxdall", uns_t, optional },
    { "trace", log_t, optional },
    { NULL }
  };

  /* Clean my variables. */
  memset ((char *)&dv, 0, sizeof(dv));
  dv.dv_rrst = 1;

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("dquote", dv.dv_dquote);
  get_arg ("bovh", dv.dv_bovh);
  get_arg ("stinframe", dv.dv_stinframe);
  get_arg ("stovh", dv.dv_stovh);
  get_arg ("maxdall", dv.dv_maxdall);
  get_arg ("trace", trace);

  if (dv.dv_stinframe == 0)
    dv.dv_stinframe = 8;
  dv.dv_na = excess_intdiv (stno, dv.dv_stinframe);

  if (dv.dv_stovh == 0)
    return_error ("stovh must be positive");
  if (dv.dv_stovh * dv.dv_stinframe >= framesize)
    return_error ("stovh*stinframe must be less than framesize");

  if (dv.dv_bovh >= framesize - dv.dv_stovh * dv.dv_stinframe)
    return_error ("bovh must be less than framesize minus overhead");

  if (dv.dv_dquote <= 0)
    return_error ("dquote must be positive");

  if (dv.dv_maxdall == 0)
    dv.dv_maxdall = (framesize * dv.dv_na - stno * dv.dv_stovh) / (2*dv.dv_na);
  dv.dv_maxdreq = dv.dv_maxdall / dv.dv_dquote;

  dv.dv_dallocation = allocate ((stno+1) * (sizeof(*dv.dv_dallocation)));
  dv.dv_vallocation = allocate ((stno+1) * (sizeof(*dv.dv_vallocation)));

  if (cyclelen < excess_intdiv(rttime,frametime)+1+1 + 2)
    return_error ("cyclelen is too small for drifs allocator to work.");
  all_delay = excess_intdiv(rttime,frametime)+1+1;

  tot_str_tru = 0;
  for_all_stations
    tot_str_tru += stp->sreq + stp->vmaxreq;

  if (tot_str_tru + dv.dv_stovh * dv.dv_stinframe > framesize)
    return_error
     ("stream and vbrmax requests must be less than framesize minus overhead");

  return (&drifs_allocator);
} /* drifs_reader */

void
free_allocators_structures ()
{
  if (cfra.vsat != NULL)
    {
      free (cfra.vsat);
      cfra.vsat = NULL;
    }
  if (g.scb != NULL)
    {
      free (g.scb);
      g.scb = NULL;
    }
}
