/*
 * Karl Heuer, 10-May-1994.  Public domain.
 * Francesco Potorti`, 13-May-1994.  Debugged, improved and made up.
 *
 * A debugging version of malloc, calloc, realloc, and free.
 * There are three independent compile-time options:
 *
 * If CHECKMALLOC is non-zero (default), then we allocate extra space
 * around each chunk and write known values there.  On realloc or
 * free, we verify that the buffer has not been overrun.  CHECKMALLOC
 * also causes the initial contents of a malloc'd block to be
 * non-zero, to make it easier to catch code that (incorrectly)
 * assumes that malloc'd space is zero.
 *
 * If TRACEMALLOC is non-zero (default), then we log each allocate or
 * free request to a temp file named /tmp/xxxmalloc, where xxx is the
 * process id of checked program.  A separate tool can scan this file,
 * deleting the corresponding allocates and frees; the remaining lines
 * will be either unmatched allocates (memory leaks) or frees (serious
 * program error).  The lines are logged with file name and line
 * number (if chkmalloc.h was included), as well as pointer value and
 * size.  This option can be enabled/disabled if the program is run
 * under a debugger by changing the value of the `tracemalloc'
 * variable.  (Warning: In order not to interfere with the program's
 * assumptions about the state of the file descriptors, the temp file
 * is opened separately on each call.  This may have a severe impact
 * on program speed.)
 *
 * if NULLFREECHECK is non-zero (default) the the argument of free is
 * checked for a non-null value.
 *
 * If HACKMALLOC is non-zero (default is zero), then we assume that
 * the library has been hacked (probably with a binary editor) to
 * rename the standard malloc, calloc, realloc, and free to maLLoc,
 * caLLoc, reaLLoc, and frEE, respectively.  This allows us to trace
 * calls that are made from a library, in addition to those that were
 * compiled with chkmalloc.h.  */

/*
 * Pick reasonable defaults.
 * Note: to turn off a feature, use -DFOO=0, not -UFOO.
 */
#ifndef CHECKMALLOC
#define CHECKMALLOC 1
#endif
#ifndef TRACEMALLOC
#define TRACEMALLOC 1
#endif
#ifndef NULLFREECHECK
#define NULLFREECHECK 1
#endif
#ifndef HACKMALLOC
#define HACKMALLOC 0
#endif
#if !defined (const) && !defined (__STDC__) && !_X_CONST
#define const
#endif
#if HACKMALLOC
extern charp maLLoc(), caLLoc(), reaLLoc();
extern void  frEE();
#else
#define maLLoc malloc
#define caLLoc calloc
#define reaLLoc realloc
#define frEE free
#endif
#if TRACEMALLOC
extern int errno;
static char dbgbuf[256];
#endif

#include <fcntl.h>
#ifndef NULL
# define NULL 0
#endif

typedef unsigned char *charp;
extern charp malloc(), calloc(), realloc();

#if CHECKMALLOC
typedef struct junk
{
  int junk_mark;
  int junk_n;
}
junk;
#define STARTMARK  ((int)0xa5a5a5a5)
#define ENDMARK    ((int)0x5a5a5a5a)
#define JUNKBYTE   0x81
#define UPALIGN(n) ((n) + (-(n) & 7))
#define JUNKPAD(n) (UPALIGN(n) + 2*sizeof(junk))
#define J(p) ((junk *)(p))
#else
#define JUNKPAD(n) (n)
#endif


static tracemalloc = TRACEMALLOC;
static in_chkmalloc = 0;

static void 
dbgwrite ()
{
  int fd, len;
  char fname[32];
  int saverrno = errno;

  sprintf (fname, "/tmp/%dmalloc", getpid ());
  fd = open (fname, O_WRONLY|O_CREAT|O_APPEND, 0666);
  if (fd == -1)
    {
      perror ("trace_malloc (): open ()");
      abort ();
    }
  len = strlen (dbgbuf);
  if (write (fd, dbgbuf, len) != len)
    {
      perror ("trace_malloc (): write ()");
      abort ();
    }
  if (close (fd) == -1)
    {
      perror ("trace_malloc (): close ()");
      abort ();
    }
  errno = saverrno;
}



#if CHECKMALLOC
static charp
dbglimit (p, n)
     charp p;
     unsigned n;
{
  charp s;

  if (p == NULL)
    return (NULL);

  J(p)->junk_mark = STARTMARK;
  J(p)->junk_n = n;
  p += sizeof(junk);
  s = p + UPALIGN(n);
  J(s)->junk_mark = ENDMARK;
  J(s)->junk_n = -n;

  return (p);
}



static charp
dbgcheck (p)
     charp p;
{
  charp s, t;
  unsigned n;

  if (p == NULL)
    return (NULL);

  p -= sizeof(junk);
  if (J(p)->junk_mark != STARTMARK)
    {
      if (tracemalloc)
	{
	  sprintf (dbgbuf+strlen(dbgbuf),
		   " at 0x%x was 0x%.8x instead of start mark\n",
		   &J(p)->junk_mark, J(p)->junk_mark);
	  dbgwrite ();
	}
      abort ();
    }
  n = J(p)->junk_n;
  s = p + sizeof(junk) + n;
  t = p + sizeof(junk) + UPALIGN(n);
  if (J(t)->junk_mark != ENDMARK)
    {
      if (tracemalloc)
	{
	  sprintf (dbgbuf+strlen(dbgbuf),
		   " at 0x%x was 0x%.8x instead of end mark\n",
		   &J(t)->junk_mark, J(t)->junk_mark);
	  dbgwrite ();
	}
      abort ();
    }
  if (J(t)->junk_n != -n)
    {
      if (tracemalloc)
	{
	  sprintf (dbgbuf+strlen(dbgbuf),
		   " at 0x%x was %d instead of %d\n",
		   &J(t)->junk_n, J(t)->junk_n, -n);
	  dbgwrite ();
	}
      abort ();
    }
  while (s < t)
    if (*(charp)s++ != JUNKBYTE)
      {
	if (tracemalloc)
	  {
	    s--;
	    sprintf (dbgbuf+strlen(dbgbuf),
		     " at 0x%x was 0x%.2x instead of 0x%.2x\n",
		     s, s[0], JUNKBYTE);
	    dbgwrite ();
	  }
	abort ();
      }

  return (p);
}
#else
#define dbgcheck (p) (p)
#endif



charp 
trace_malloc (file, line, n)
     char const *file;
     int line; 
     unsigned n;
{
  charp p;

  if (in_chkmalloc)
    return maLLoc (n);
  in_chkmalloc += 1;
  p = maLLoc (JUNKPAD(n));

  if (CHECKMALLOC)
    {
      p = dbglimit (p, n);
      memset (p, JUNKBYTE, UPALIGN(n));
    }
  if (tracemalloc)
    {
      sprintf (dbgbuf, "%s:%d: malloc(%u) -> 0x%x\n", file, line, n, p);
      dbgwrite ();
    }
  in_chkmalloc -= 1;
  return (p);
}



charp
trace_calloc (file, line, m, k)
     char const *file; 
     int line;
     unsigned m, k;
{
  unsigned n;
  charp p;

  if (in_chkmalloc)
    return caLLoc (m, k);
  in_chkmalloc += 1;
  n = m*k;
  p = caLLoc (JUNKPAD(n), 1);

  if (CHECKMALLOC)
    {
      p = dbglimit (p, n);
      memset (p+n, JUNKBYTE, UPALIGN(n)-n);
    }
  if (tracemalloc)
    {
      sprintf (dbgbuf, "%s:%d: calloc(%u, %u) -> 0x%x\n",
	       file, line, m, k, p);
      dbgwrite ();
    }
  in_chkmalloc -= 1;
  return (p);
}



charp
trace_realloc (file, line, oldp, n) 
     char const *file;
     int line;
     charp oldp;
     unsigned n;
{
  charp p;
  unsigned oldn;

  if (in_chkmalloc)
    return reaLLoc (oldp, n);
  in_chkmalloc += 1;

  if (tracemalloc)
    sprintf (dbgbuf, "%s:%d: realloc(0x%x, %u)", file, line, oldp, n);

  p = reaLLoc (dbgcheck (oldp), JUNKPAD(n));

  if (CHECKMALLOC)
    {
      oldn = J(p)->junk_n;
      p = dbglimit (p, n);
      if (n > oldn)
	memset (p+oldn, JUNKBYTE, UPALIGN(n)-oldn);
      else
	memset (p+n, JUNKBYTE, UPALIGN(n)-n);
    }
  if (tracemalloc)
    {
      sprintf (dbgbuf+strlen(dbgbuf), " -> 0x%x\n", p);
      dbgwrite ();
    }
  in_chkmalloc -= 1;
  return (p);
}



void 
trace_free (file, line, p)
     char const *file;
     int line;
     charp p;
{
  if (in_chkmalloc)
    {
      frEE (p);
      return;
    }
  in_chkmalloc += 1;

  if (tracemalloc)
    sprintf (dbgbuf, "%s:%d: free(0x%x)", file, line, p);
  if (NULLFREECHECK && p == 0)
    {
      if (tracemalloc)
	{
	  sprintf (dbgbuf+strlen(dbgbuf), " called with null argument\n");
	  dbgwrite ();
	}
      abort ();
    }
  frEE (dbgcheck (p));
  if (tracemalloc)
    {
      strcat (dbgbuf, "\n");
      dbgwrite ();
    }
  in_chkmalloc -= 1;
}

#if HACKMALLOC
charp
malloc (n)
     unsigned n;
{
  return (trace_malloc ("library", 0, n));
}

charp
calloc (m, n)
     unsigned m, n;
{
  return (trace_calloc ("library", 0, m, n));
}

charp
realloc (oldp, n)
     charp oldp;
     unsigned n;
{
  return (trace_realloc ("library", 0, oldp, n));
}

void
free (p)
     charp p;
{
  trace_free ("library", 0, p);
}
#endif
