/****************************************************************************

  $Source: /home/foda/fracas/RCS/generators.c $
  $Revision: 7.11 $
  $Date: 1998/11/25 11:38:40 $
  $State: Exp $
  $Locker:  $

****************************************************************************/
#ident "$Id: generators.c 7.11 1998/11/25 11:38:40 pot Exp $"

/*
 *  FRACAS stands for FRAmed Channel Access Simulator.
 */

#include <stdio.h>
#include "fracas.h"
#include "libnr.h"

static sampler_r constant_reader;
static sampler_r poisson_reader;
static sampler_r impulse_reader;
static sampler_r enst_bulk_reader;
static sampler_r enst_interactive_reader;
static sampler_r fgn_reader;
static sampler_r external_reader;
static sampler_r markov_reader;

sampler_n samplers [] =
{
  { constant_reader, "constant" },
  { poisson_reader, "poisson" },
  { impulse_reader, "impulse" },
  { enst_bulk_reader, "enst-bulk" },
  { enst_interactive_reader, "enst-interactive" },
  { fgn_reader, "fgn" },
  { external_reader, "external" },
  { markov_reader, "markov" },
  { NULL }
};

static int
null_sampler (tg_args *argsp)
{
  return (0);
}


/*
 * Constant sampler.
 * Keeps state information.
 * Generates TRUs at fixed time intervals.
 *
 * Syntax is: constant traffic|tfactor [start] [stop] [cycle] [burst]
 *	being <traffic> the floating point number of TRUs per TMU.
 *	The <start> optional floating point parameter is the time
 *	when the generator starts transmitting (default 0).
 *	The <stop> optional floating point parameter is the time
 *	when the generator stops transmitting, 0 meaning never (default 0).
 *	The <cycle> optional floating point parameter means that the
 *	start-stop cycle should be cyclically reset after <cycle> time
 *	units, 0 meaning an infinite cycle length (default 0).
 */
static int
constant_sampler (tg_args *argsp)
{
  struct constant_args *tap = &argsp->constant;
  struct cycle_vars *cvp = &tap->cv;
  float interval;
  int trus;

  interval = tap->interval;
  if (cvp->on)
    {
      /* The generator is on. */
      float next_off_switch = cvp->next_off_switch;

      if (next_off_switch > 0 && next_off_switch < curtime + frametime)
	{
	  /* The generator will stop during this frame. */
	  float cycle = cvp->cycle;

	  if (cycle > 0)
	    cvp->next_off_switch += cycle;
	  else
	    cvp->next_off_switch = 0;
	  cvp->on = FALSE;

	  trus = (next_off_switch - tap->last_gen) / interval;
	}
      else
	{
	  /* The generator will not stop during this frame. */
	  trus = (curtime + frametime - tap->last_gen) / interval;
	  if (trus > 0)
	    tap->last_gen += trus * interval;
	}
    }
  else
    {
      /* The generator is off. */
      float next_on_switch = cvp->next_on_switch;
      if (next_on_switch > 0 && next_on_switch < curtime + frametime)
	{
	  /* The generator will start during this frame. */
	  float cycle = cvp->cycle;

	  if (cycle > 0)
	    cvp->next_on_switch += cycle;
	  else
	    cvp->next_on_switch = 0;
	  cvp->on = TRUE;

	  trus = (curtime + frametime - next_on_switch) / interval + 1;
	  tap->last_gen = next_on_switch + (trus -1) * interval;
	}
      else
	{
	  /* The generator will not restart during this frame. */
	  trus = 0;
	}
    }

  return (trus * tap->burst);
} /* constant_sampler() */


static int
check_and_set_cycle_vars (struct cycle_vars *cvp)
{
  if (cvp->stop < 0)
    return_error ("stop must not be negative");
  if (cvp->cycle > 0
      && cvp->start > ((cvp->stop == 0) ? cvp->cycle : cvp->stop))
    return_error ("start must not be greater than stop");
  if (cvp->stop > 0 && cvp->stop - cvp->start < frametime)
    return_error ("on time must not be less than frametime");

  if (cvp->cycle < 0)
    return_error ("cycle must not be negative");
  if (cvp->cycle > 0 && cvp->cycle < cvp->stop)
    return_error ("cycle must not be less than stop");
  if (cvp->cycle > 0 && cvp->start < 0 && cvp->stop - cvp->start > cvp->cycle)
    return_error ("on time must not be greater than cycle");
  if (cvp->cycle > 0 && cvp->stop == 0 && cvp->cycle - cvp->start < frametime)
    return_error ("on time must not be less than frametime");
  if (cvp->cycle > 0 && ((cvp->stop == 0)
			 ? cvp->start
			 : cvp->start + cvp->cycle - cvp->stop) < frametime)
    return_error ("off time must not be less than frametime");

  cvp->on = (cvp->start <= 0);
  if (cvp->start < 0)
    cvp->next_on_switch = cvp->start + cvp->cycle;
  else
    cvp->next_on_switch = cvp->start;
  if (cvp->stop > 0)
    cvp->next_off_switch = cvp->stop;
  else
    cvp->next_off_switch = cvp->cycle;

  return (1);			/* success */
}

tg *
constant_reader (char *lp)
{
  float traffic, tfactor;

  char *errstr;
  tg *tgp = allocate (sizeof (tg));
  struct constant_args *tap = &tgp->t_args.constant;
  arg_type args[] =
    {
      { "traffic", float_t, optional },
      { "tfactor", float_t, optional },
      { "start", float_t, optional },
      { "stop", float_t, optional },
      { "cycle", float_t, optional },
      { "burst", uns_t, optional },
      { NULL }
    };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);

  get_arg ("traffic", traffic);
  get_arg ("tfactor", tfactor);
  if (traffic == 0)
    {
      if (tfactor == 0)
	{
	  input_warning ("generator has null traffic");
	  tgp->t_samplerp = &null_sampler;
	  return (tgp);
	}
      if (tfactor < 0)
	return_error ("tfactor must be positive");
      if (reference_traffic == 0)
	return_error ("tfactor requires ref_traffic to be specified");
      traffic = tfactor * reference_traffic;
    }
  else if (traffic < 0)
    return_error ("traffic must be positive");
  else if (tfactor != 0)
    return_error ("only one of traffic and tfactor can be specified");

  tap->interval = 1/traffic;

  get_arg ("start", tap->cv.start);
  get_arg ("stop", tap->cv.stop);
  get_arg ("cycle", tap->cv.cycle);
  if (check_and_set_cycle_vars (&tap->cv) == 0)
    return (NULL);

  get_arg ("burst", tap->burst);
  if (tap->burst == 0)
    tap->burst = 1;

  tap->last_gen = 0;
  tgp->t_samplerp = &constant_sampler;

  return (tgp);
} /* constant_reader() */

/*
 * Poisson sampler.
 * Stateless.
 * Generates TRUs at exponential time intervals.
 *
 * Syntax is: poisson traffic|tfactor [start] [stop] [cycle] [seed] [burst]
 *	being <traffic> the floating point mean number of TRUs per TMU.
 *	The <start> optional floating point parameter is the time
 *	when the generator starts transmitting (default 0).
 *	The <stop> optional floating point parameter is the time
 *	when the generator stops transmitting, 0 meaning never (default 0).
 *	The <cycle> optional floating point parameter means that the
 *	start-stop cycle should be cyclically reset after <cycle> time
 *	units, 0 meaning an infinite cycle length (default 0).
 */
static int
poisson_sampler (tg_args *argsp)
{
  struct poisson_args *tap = &argsp->poisson;
  struct cycle_vars *cvp = &tap->cv;
  double LambdaExp = tap->LambdaExp;
  float r;
  double t;
  float ontime;
  int trus;

  setstate (&tap->rs);

  /* Compute the length of the on time. */
  if (cvp->on)
    {
      /* The generator is on. */
      float next_off_switch = cvp->next_off_switch;

      if (next_off_switch > 0 && next_off_switch < curtime + frametime)
	{
	  /* The generator will stop during this frame. */
	  float cycle = cvp->cycle;

	  if (cycle > 0)
	    cvp->next_off_switch += cycle;
	  else
	    cvp->next_off_switch = 0;
	  cvp->on = FALSE;

	  ontime = next_off_switch - curtime;
	}
      else
	{
	  /* The generator will not stop during this frame. */
	  ontime = frametime;
	}
    }
  else
    {
      /* The generator is off. */
      float next_on_switch = cvp->next_on_switch;
      if (next_on_switch > 0 && next_on_switch < curtime + frametime)
	{
	  /* The generator will start during this frame. */
	  float cycle = cvp->cycle;

	  if (cycle > 0)
	    cvp->next_on_switch += cycle;
	  else
	    cvp->next_on_switch = 0;
	  cvp->on = TRUE;

	  ontime = curtime + frametime - next_on_switch;
	}
      else
	{
	  /* The generator will not restart during this frame. */
	  ontime = 0;
	}
    }

  /* This implementation takes a time proportional to lambda, on
     average.  See Numerical recipes in C, pg.294 for an algorithm
     that is efficient with high lambda's. */

  /* Instead of summing exponentially distributed times until they
     exceed a threshold, which would require a log of every random
     sample, we multiply uniformly distributed samples and compare
     the product with the exponential of the threshold. */
  trus = 0;
  if (ontime > 0)
    {
      if (ontime != frametime)
	LambdaExp = exp (-tap->lambda * ontime/frametime);
      do r = frandom(); while (r == 0);
      t = r;
      while (t > LambdaExp)
	{
	  trus += 1;
	  do r = frandom(); while (r == 0);
	  t *= r;
	}
    }

  return (trus * tap->burst);
} /* poisson_sampler() */

tg *
poisson_reader (char *lp)
{
  float lambda, traffic, tfactor;
  int seed;

  char *errstr;
  tg *tgp = allocate (sizeof (tg));
  struct poisson_args *tap = &tgp->t_args.poisson;
  arg_type args[] =
    {
      { "traffic", float_t, optional },
      { "tfactor", float_t, optional },
      { "start", float_t, optional },
      { "stop", float_t, optional },
      { "cycle", float_t, optional },
      { "seed", int_t, optional },
      { "burst", uns_t, optional },
      { NULL }
    };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("traffic", traffic);
  get_arg ("tfactor", tfactor);
  get_arg ("seed", seed);

  if (seed == 0)
    seed = random ();

  if (traffic == 0)
    {
      if (tfactor == 0)
	{
	  input_warning ("generator has null traffic");
	  tgp->t_samplerp = &null_sampler;
	  return (tgp);
	}
      if (tfactor < 0)
	return_error ("tfactor must be positive");
      if (reference_traffic == 0)
	return_error ("tfactor requires ref_traffic to be specified");
      traffic = tfactor * reference_traffic;
    }
  else if (traffic < 0)
    return_error ("traffic must be positive");
  else if (tfactor != 0)
    return_error ("only one of traffic and tfactor can be specified");

  lambda = traffic * frametime;
  if (lambda > -log (DBL_MIN))
    {
      char errbuf [200];
      sprintf (errbuf, "traffic is too great by %.1f times",
	       lambda / -log (DBL_MIN));
      return_error (errbuf);
    }

  get_arg ("burst", tap->burst);
  if (tap->burst == 0)
    tap->burst = 1;

  get_arg ("start", tap->cv.start);
  get_arg ("stop", tap->cv.stop);
  get_arg ("cycle", tap->cv.cycle);
  if (check_and_set_cycle_vars (&tap->cv) == 0)
    return (NULL);

  tap->lambda = lambda;
  tap->LambdaExp = exp (-lambda);
  setstate (initstate (seed, &tap->rs, RND_STATE_SIZE));
  tgp->t_samplerp = &poisson_sampler;

  return (tgp);
} /* poisson_reader() */

/*
 * Two-state markov modulated poisson sampler.
 * Keeps state information.
 * Generates TRUs at markov modulated exponential interval times.
 *
 * Syntax is: impulse traffic|tfactor cycle duty burstiness [seed] [burst]
 *	being <traffic> the mean number of TRUs per TMU, <cycle>
 * 	the mean cycle length, <duty> the mean ratio of length
 *  	of state 1 to cycle length and <burstiness> the mean
 *  	ratio of traffic generated in state 1 to the mean traffic
 *   	<traffic>.  All are positive floating point numbers, <duty> is
 *    	less than 1, and <burstiness>*<duty> is less than 1.
 */
static int
impulse_sampler (tg_args *argsp)
{
  struct impulse_args *tap = &argsp->impulse;
  int trus, state;
  float timet, time1;
  double LambdaExp;
  float r, rlog;
  double t;

  setstate (&tap->rs);

  /* The algorithm for computing how much time we spend in each state
     is efficient only if the transitions are not very frequent in a
     frame (which is the common case).  The number of TRUs in each
     state is computed with an algorithm that needs a time proportional
     to  the number of TRUs generated: see the comment for the poisson
     generator. */

  /* First we compute how much time we stay in each of the two states.
     We compute the time we stay in state 1 (time1) and the total time
     (timet) until it exceeds a frame. */
  state = tap->state;
  timet = 0;
  time1 = 0;
  do
    {
      do r = frandom(); while (r == 0);
      rlog = -log(r);
      switch (state)
	{
	case 1:
	  rlog *= tap->u11;
	  timet += rlog;
	  time1 += rlog;
	  if (timet < frametime)
	    state = 2;
	  else
	    time1 -= timet - frametime;
	  break;
	case 2:
	  rlog *= tap->u22;
	  timet += rlog;
	  if (timet < frametime)
	    state = 1;
	  break;
	} /* switch (state) */
    }
  while (timet < frametime);
  tap->state = state;

  /* Compute the number of TRUs generated in each of the two states. */
  trus = 0;
  if (time1 > 0)
    /* TRUs generated in state 1 */
    {
      if (time1 == frametime)
	LambdaExp = tap->Lambda1Exp;
      else
	LambdaExp = exp(-time1/tap->interval1);
      do r = frandom(); while (r == 0);
      t = r;
      while (t > LambdaExp)
	{
	  trus += 1;
	  do r = frandom(); while (r == 0);
	  t *= r;
	}
    }
  if (time1 < frametime)
    /* TRUs generated in state 2 */
    {
      if (time1 == 0)
	LambdaExp = tap->Lambda2Exp;
      else
	LambdaExp = exp((time1-frametime)/tap->interval2);
      do r = frandom(); while (r == 0);
      t = r;
      while (t > LambdaExp)
	{
	  trus += 1;
	  do r = frandom(); while (r == 0);
	  t *= r;
	}
    }

  return (trus * tap->burst);
} /* impulse_sampler() */

tg *
impulse_reader (char *lp)
{
  float traffic, tfactor, cyclelength, dutycycle, burstiness;
  float Lambda1, Lambda2;
  random_state *rsp;
  int seed;

  char *errstr;
  tg *tgp = allocate (sizeof (tg));
  struct impulse_args *tap = &tgp->t_args.impulse;
  arg_type args[] =
    {
      { "traffic", float_t, optional },
      { "tfactor", float_t, optional },
      { "cycle", float_t, required },
      { "duty", float_t, required },
      { "burstiness", float_t, required },
      { "seed", int_t, optional },
      { "burst", uns_t, optional },
      { NULL }
    };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("traffic", traffic);
  get_arg ("tfactor", tfactor);
  get_arg ("cycle", cyclelength);
  get_arg ("duty", dutycycle);
  get_arg ("burstiness", burstiness);
  get_arg ("seed", seed);

  if (seed == 0)
    seed = random ();

  get_arg ("burst", tap->burst);
  if (tap->burst == 0)
    tap->burst = 1;

  if (traffic == 0)
    {
      if (tfactor == 0)
	{
	  input_warning ("generator has null traffic");
	  tgp->t_samplerp = &null_sampler;
	  return (tgp);
	}
      if (tfactor < 0)
	return_error ("tfactor must be positive");
      if (reference_traffic == 0)
	return_error ("tfactor requires ref_traffic to be specified");
      traffic = tfactor * reference_traffic;
    }
  else if (traffic < 0)
    return_error ("traffic must be positive");
  else if (tfactor != 0)
    return_error ("only one of traffic and tfactor can be specified");

  if (cyclelength <= 0)
    return_error ("cycle must be positive");
  if (dutycycle <= 0 || dutycycle >= 1)
    return_error ("duty must be positive and less than 1");
  if (burstiness <= 0)
    return_error ("burstiness must be positive");
  if (burstiness*dutycycle >= 1)
    return_error ("(burstiness*duty) must be less than 1");

  Lambda1 = traffic * burstiness * frametime;
  if (Lambda1 > -log (DBL_MIN))
    {
      char errbuf [200];
      sprintf (errbuf,
	       "generation rate in state 1 is too big by %.1f times",
	       Lambda1 / -log (DBL_MIN));
      return_error (errbuf);
    }

  Lambda2 = traffic * (1-burstiness*dutycycle)/(1-dutycycle) * frametime;
  if (Lambda2 > -log (DBL_MIN))
    {
      char errbuf [200];
      sprintf (errbuf,
	       "generation rate in state 2 is too big by %.1f times",
	       Lambda2 / -log (DBL_MIN));
      return_error (errbuf);
    }

  tap->interval1 = frametime / Lambda1;
  tap->Lambda1Exp = exp (-Lambda1);
  tap->interval2 = frametime / Lambda2;
  tap->Lambda2Exp = exp (-Lambda2);
  tap->u11 = cyclelength * dutycycle;
  tap->u22 = cyclelength * (1-dutycycle);
  rsp = initstate (seed, &tap->rs, RND_STATE_SIZE);
  tap->state = (frandom() < dutycycle) ? 1 : 2;
  setstate (rsp);
  tgp->t_samplerp = &impulse_sampler;

  return (tgp);
} /* impulse_reader() */

/*
 * ENST bulk generator sampler.
 * Keeps state information.
 * Generates TRUs as the result of file transfers.
 *  Files are generated at exponentially distributed interval times.
 *  The length of the files is chosen according to a constant plus
 *  exponential distribution.  The files are sent with a rate chosen
 *  among two possible ones.  The choice between the two depends on
 *  the state of a two-state markov chain.  The constant rate state
 *  sends packets at fixed interval times.  The exponential state
 *  sends packet with interval time chosen according to a constant
 *  plus exponential distribution.
 *
 * Syntax is: enst-bulk traffic|tfactor minlen meanlen consttime constrate
 *	                [exptime] [maxrate] [exprate] [seed] [burst]
 *	being <traffic> the mean number of TRUs per TMU, <minlen>
 * 	the minimum length of a generated file (that is, the minimum
 * 	number of TRUs generated when a file is to be sent), <meanlen>
 * 	the mean file length.  The length of a file is the constant
 * 	<minlen> plus an exponentially distributed quantity whose mean
 * 	is <meanlen> - <minlen>.  The files are generated at intervals
 *  	following an exponential distribution whose mean is such that
 *  	the overall throughput equals <traffic>.  The files are
 *   	packetised in TRUs: the number of TRUs generated by a file is
 *   	the smallest integer not greater than the file length.  The
 *    	rate at which the file data are packetised into TRUs depends
 *     	on the state of the markov process.  <consttime> is the mean
 *     	time of permanence in state 1, where TRUs are generated a
 *     	constant rate <constrate>.  <exptime> (default <consttime>) is
 *     	the mean time of permanence in state 2, where TRUs are
 *     	generated with interval times equal to a minimum 1/<maxrate>
 *     	(default <constrate>) plus an exponential distributed time
 *     	such that the mean rate in state 2 is <exprate> (default
 *     	<constrate>/4). All the parameters are positive floating point
 *     	numbers.
 */
static int
enst_bulk_sampler (tg_args *argsp)
{
  struct enst_bulk_args *tap = &argsp->enst_bulk;
  int trus, state;
  float timet, time1, time2;
  float r, rlog;
  double t;

  setstate (&tap->rs);

  /* The algorithm for computing how much time we spend in each state
     is efficient only if the transitions are not very frequent in a
     frame (which is the common case).  The number of TRUs in state 2
     is computed with an algorithm that needs a time increasing with
     the number of TRUs generated: see the comment for the poisson
     generator.  The computation of the number of files generated has
     the same efficiency problem. */

  /* First we compute the total number of TRUs obtained by
     packetising the files generated during this frame and add it to
     the backlog. */
  do r = frandom(); while (r == 0);
  t = r;
  while (t > tap->LambdaFiletimeExp)
    {
      do r = frandom(); while (r == 0);
      tap->backlog += ceil (tap->fileminlen - tap->filelen*log(r));
      do r = frandom(); while (r == 0);
      t *= r;
    }


  /* Second we compute how much time we stay in each of the two states.
     We compute the time we stay in state 1 (time1) and the total time
     (timet) until it exceeds a frame. */
  state = tap->state;
  timet = 0;
  time1 = 0;
  do
    {
      do r = frandom(); while (r == 0);
      rlog = -log(r);
      switch (state)
	{
	case 1:
	  rlog *= tap->u11;
	  timet += rlog;
	  time1 += rlog;
	  if (timet < frametime)
	    state = 2;
	  else
	    time1 -= timet - frametime;
	  break;
	case 2:
	  rlog *= tap->u22;
	  timet += rlog;
	  if (timet < frametime)
	    state = 1;
	  break;
	} /* switch (state) */
    }
  while (timet < frametime);
  tap->state = state;
  if (tap->backlog == 0)
    return (0);

  /* Next we compute the number of TRUs generated in each of the two
     states.  We start with state 1, which is quicker to compute
     because the interarrival times are constant, so if the backlog is
     done with in this state we need not make computations for state
     2, where the interarrival times are constant + exponential. */
  trus = 0;
  if (time1 > 0)
    /* TRUs generated in state 1 (constant rate) */
    {
      trus += 0.5 + time1 / tap->constinterval;
      if (trus > tap->backlog)
	trus = tap->backlog;
      tap->backlog -= trus;
    }
  time2 = frametime - time1;
  if (time2 > 0 && tap->backlog > 0)
    /* TRUs generated in state 2 */
    {
      do r = frandom(); while (r == 0);
      rlog = -log(r);
      while (tap->backlog > 0
	     && (time2 -= tap->mininterval + rlog*tap->interval2) > 0)
	{
	  trus += 1;
	  tap->backlog -= 1;
	  do r = frandom(); while (r == 0);
	  rlog = -log(r);
	}
    }

  return (trus * tap->burst);
} /* enst_bulk_sampler() */

tg *
enst_bulk_reader (char *lp)
{
  float traffic, tfactor;
  float minlen, meanlen, consttime, constrate, exptime, maxrate, exprate;
  float LambdaFiletime, invfilelen, Lambda2;
  double p, E;
  random_state *rsp;
  int seed;

  char *errstr;
  tg *tgp = allocate (sizeof (tg));
  struct enst_bulk_args *tap = &tgp->t_args.enst_bulk;
  arg_type args[] =
    {
      { "traffic", float_t, optional },
      { "tfactor", float_t, optional },
      { "minlen", float_t, required },
      { "meanlen", float_t, required },
      { "consttime", float_t, required },
      { "constrate", float_t, required },
      { "exptime", float_t, optional },
      { "maxrate", float_t, optional },
      { "exprate", float_t, optional },
      { "seed", int_t, optional },
      { "burst", uns_t, optional },
      { NULL }
    };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("traffic", traffic);
  get_arg ("tfactor", tfactor);
  get_arg ("minlen", minlen);
  get_arg ("meanlen", meanlen);
  get_arg ("consttime", consttime);
  get_arg ("constrate", constrate);
  get_arg ("exptime", exptime);
  get_arg ("maxrate", maxrate);
  get_arg ("exprate", exprate);
  get_arg ("seed", seed);

  if (seed == 0)
    seed = random ();

  get_arg ("burst", tap->burst);
  if (tap->burst == 0)
    tap->burst = 1;

  if (traffic == 0)
    {
      if (tfactor == 0)
	{
	  input_warning ("generator has null traffic");
	  tgp->t_samplerp = &null_sampler;
	  return (tgp);
	}
      if (tfactor < 0)
	return_error ("tfactor must be positive");
      if (reference_traffic == 0)
	return_error ("tfactor requires ref_traffic to be specified");
      traffic = tfactor * reference_traffic;
    }
  else if (traffic < 0)
    return_error ("traffic must be positive");
  else if (tfactor != 0)
    return_error ("only one of traffic and tfactor can be specified");

  get_arg ("maxrate", maxrate);
  get_arg ("exprate", exprate);
  if (minlen <= 0)
    return_error ("minlen must be positive");
  if (meanlen <= 0 || meanlen <= minlen)
    return_error ("meanlen must be positive and greater than minlen");
  if (consttime <= 0)
    return_error ("consttime must be positive");
  if (constrate <= 0)
    return_error ("constrate must be positive");
  if (exptime == 0)
    exptime = consttime;
  else if (exptime < 0)
    return_error ("exptime must be positive");
  if (maxrate == 0)
    maxrate = constrate;
  else if (maxrate <= 0)
    return_error ("maxrate must be positive");
  if (exprate == 0)
    exprate = constrate / 4;
  else if (exprate <= 0 || exprate > maxrate)
    return_error ("exprate must be positive and not greater than maxrate");

  /* Check that the mean rate of the packetiser is at least 10%
     greater than that of the file generator. */
  if (traffic
      > 0.9 * (exptime*exprate + consttime*constrate) / (exptime+consttime))
    input_warning ("file generator rate higher than 90% of packetiser's");

  /* Compute the mean of the exponential distribution of the file
     intergeneration time.  It has to be corrected with a factor that
     accounts for the packetisation process.  Indeed, the mean file
     lenght is (meanlen) only before packetisation.  After
     packetisation, it is (p + E^(p-minlen)/(1-E)).  Look at the code
     for the meaning of p and E. */
  /* This one is only approximate
  LambdaFiletime = traffic * frametime / meanlen; */
  p = ceil (minlen);
  E = exp (-1/(meanlen-minlen));
  LambdaFiletime = traffic * frametime / (p + pow(E, p-minlen)/(1-E));
  if (LambdaFiletime > -log (DBL_MIN))
    {
      char errbuf [200];
      sprintf (errbuf,
	       "file generation rate is too big by %.1f times",
	       LambdaFiletime / -log (DBL_MIN));
      return_error (errbuf);
    }

  invfilelen = 1 / (meanlen - minlen);
  if (invfilelen > -log (DBL_MIN))
    {
      char errbuf [200];
      sprintf (errbuf,
	       "meanlen - minlen is too small by %.1f times",
	       invfilelen / -log (DBL_MIN));
      return_error (errbuf);
    }

  Lambda2 = frametime / (1/exprate - 1/maxrate);
  if (Lambda2 > -log (DBL_MIN))
    {
      char errbuf [200];
      sprintf (errbuf,
	       "maxrate - meanrate is too small by %.1f times",
	       Lambda2 / -log (DBL_MIN));
      return_error (errbuf);
    }

  tap->LambdaFiletimeExp = exp (-LambdaFiletime);
  tap->fileminlen = minlen;
  tap->filelen = 1 / invfilelen;
  tap->invfilelenExp = exp (-invfilelen);
  tap->constinterval = 1 / constrate;
  tap->mininterval = 1 / maxrate;
  tap->interval2 = frametime / Lambda2;
  tap->u11 = consttime;
  tap->u22 = exptime;
  rsp = initstate (seed, &tap->rs, RND_STATE_SIZE);
  tap->state = (frandom() < consttime/(consttime+exptime)) ? 1 : 2;
  setstate (rsp);
  tgp->t_samplerp = &enst_bulk_sampler;

  return (tgp);
} /* enst_bulk_reader() */

/*
 * ENST interactive generator sampler.
 * Keeps state information.
 * Generates TRUs as the result of interactive traffic.
 *  Packets are generated with intergeneration times having a constant
 *  plus exponential distribution.  Their length is chosen according
 *  to a fixed table of lengths.  This table has been experimentally
 *  chosen and refers to a TRU of 10 bytes.
 *
 * Syntax is: enst-interactive traffic|tfactor mintime [seed] [burst]
 *	being <traffic> the mean number of TRUs per TMU and
 *	<mintime> is the minimum intergeneration time between packets.
 *     	All the parameters are positive floating point numbers.
 */
static int
enst_interactive_sampler (tg_args *argsp)
{
  struct enst_interactive_args *tap = &argsp->enst_interactive;
  int trus, packets;
  float r, rlog;
  double t;

  setstate (&tap->rs);

  /* The number of packets generated is computed with an algorithm
     that needs a time increasing with the number of packets
     generated: see the comment for the poisson generator. */

  packets = 0;
  for (t = tap->firstinterval - frametime;
       t < 0;
       t += tap->mininterval + rlog*tap->expinterval)
    {
      packets += 1;
      do r = frandom(); while (r == 0);
      rlog = -log(r);
    }
  tap->firstinterval = t;

  trus = 0;
  while (packets-- > 0)
    {
      int j = tap->pld_bins;
      r = frandom ();
      while (--j > 0)
	if (r >= tap->len_distrib[j].cumdistr)
	  break;
      trus += tap->len_distrib[j].packet_len;
    }

  return (trus * tap->burst);
} /* enst_interactive_sampler() */

tg *
enst_interactive_reader (char *lp)
{
  float traffic, tfactor;
  float mintime, meanlen, right_bin_margin;
  random_state *rsp;
  int j, seed;

  char *errstr;
  tg *tgp = allocate (sizeof (tg));
  struct enst_interactive_args *tap = &tgp->t_args.enst_interactive;
  arg_type args[] =
    {
      { "traffic", float_t, optional },
      { "tfactor", float_t, optional },
      { "mintime", float_t, required },
      { "seed", int_t, optional },
      { "burst", uns_t, optional },
      { NULL }
    };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);
  get_arg ("traffic", traffic);
  get_arg ("tfactor", tfactor);
  get_arg ("mintime", mintime);
  get_arg ("seed", seed);

  if (seed == 0)
    seed = random ();

  get_arg ("burst", tap->burst);
  if (tap->burst == 0)
    tap->burst = 1;

  if (traffic == 0)
    {
      if (tfactor == 0)
	{
	  input_warning ("generator has null traffic");
	  tgp->t_samplerp = &null_sampler;
	  return (tgp);
	}
      if (tfactor < 0)
	return_error ("tfactor must be positive");
      if (reference_traffic == 0)
	return_error ("tfactor requires ref_traffic to be specified");
      traffic = tfactor * reference_traffic;
    }
  else if (traffic < 0)
    return_error ("traffic must be positive");
  else if (tfactor != 0)
    return_error ("only one of traffic and tfactor can be specified");

  if (mintime <= 0)
    return_error ("mintime must be positive");

  tap->pld_bins = 5;
  tap->len_distrib = allocate (tap->pld_bins * sizeof (*tap->len_distrib));
  tap->len_distrib[0].cumdistr = 0;
  tap->len_distrib[0].packet_len = 1;
  tap->len_distrib[1].cumdistr = 0.8;
  tap->len_distrib[1].packet_len = 5;
  tap->len_distrib[2].cumdistr = 0.85;
  tap->len_distrib[2].packet_len = 10;
  tap->len_distrib[3].cumdistr = 0.9;
  tap->len_distrib[3].packet_len = 100;
  tap->len_distrib[4].cumdistr = 0.95;
  tap->len_distrib[4].packet_len = 200;

  meanlen = 0;
  right_bin_margin = 1;
  for (j = tap->pld_bins-1; j >= 0; j--)
    {
      meanlen += (tap->len_distrib[j].packet_len
		  * (right_bin_margin - tap->len_distrib[j].cumdistr));
      right_bin_margin = tap->len_distrib[j].cumdistr;
    }
  tap->firstinterval = 0;
  tap->mininterval = mintime;
  tap->expinterval = meanlen / traffic - mintime;
  rsp = initstate (seed, &tap->rs, RND_STATE_SIZE);
  setstate (rsp);
  tgp->t_samplerp = &enst_interactive_sampler;

  return (tgp);
} /* enst_interactive_reader() */

/*
 * Fractional Gaussian Noise sampler.
 * Keeps state info
 * Generates fractional gaussian noise using the Random Midpoint
 * Displacement algorithm.
 *
 * Syntax is: fgn traffic|tfactor peakedness H [rmdn] [seed] [burst]
 *	where <traffic> is the mean number of TRUs generated per TMU,
 *	<peakedness> is the ratio of variance to mean of the number of
 *	TRUs per TMU, and <H> is the hurst parameter.  All are positive
 *	floating point numbers, <H> is in the [0.5; 1) interval.
 *
 *	The optional parameter rmdn is the base-2 log of the length of
 *	a trace; it must be greater than 10, and defaults to 18.
 */
static int
fgn_sampler (tg_args *argsp)
{
  struct fgn_args *tap = &argsp->fgn;

  if (tap->samplep > tap->lastp)
    {
      setstate (&tap->rs);
      RMDtraffic (tap->average, tap->peakedness, tap->H, frametime<<tap->rmdn,
		  tap->rmdn, tap->firstp);
      tap->samplep = tap->firstp;
    }

  return ((int)(tap->burst * (*tap->samplep++) + 0.5));
} /* fgn_sampler() */

tg *
fgn_reader (char *lp)
{
  float traffic, tfactor;
  int len;
  int seed;

  char *errstr;
  tg *tgp = allocate (sizeof (tg));
  struct fgn_args *tap = &tgp->t_args.fgn;
  arg_type args[] =
    {
      { "traffic", float_t, optional },
      { "tfactor", float_t, optional },
      { "peakedness", float_t, required },
      { "H", float_t, required },
      { "rmdn", int_t, optional },
      { "seed", int_t, optional },
      { "burst", uns_t, optional },
      { NULL }
    };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);

  get_arg ("traffic", traffic);
  get_arg ("tfactor", tfactor);
  get_arg ("peakedness", tap->peakedness);
  get_arg ("H", tap->H);
  get_arg ("rmdn", tap->rmdn);
  get_arg ("burst", tap->burst);
  get_arg ("seed", seed);

  if (seed == 0)
    seed = random ();

  if (tap->burst == 0)
   tap->burst = 1;

  if (traffic == 0)
    {
      if (tfactor == 0)
	{
	  input_warning ("generator has null traffic");
	  tgp->t_samplerp = &null_sampler;
	  return (tgp);
	}
      if (tfactor < 0)
	return_error ("tfactor must be positive");
      if (reference_traffic == 0)
	return_error ("tfactor requires ref_traffic to be specified");
      traffic = tfactor * reference_traffic;
    }
  else if (traffic < 0)
    return_error ("traffic must be positive");
  else if (tfactor != 0)
    return_error ("only one of traffic and tfactor can be specified");

  tap->average = traffic;
  if (tap->peakedness <= 0)
    return_error ("peakedness must be positive");
  if (tap->H < 0.5 || tap->H >= 1)
    return_error ("H must lay in the [0.5; 1) interval");
  if (tap->rmdn == 0)
    tap->rmdn = 18;
  else if (tap->rmdn < 10 || tap->rmdn >= 30)
    return_error ("rmdn must be greater than 10 and less than 30");

  len = 1 << tap->rmdn;
  tap->firstp = allocate (sizeof(*tap->firstp) * len);
  tap->lastp = tap->firstp + len - 1;
  tap->samplep = tap->lastp + 1;
  setstate (initstate (seed, &tap->rs, RND_STATE_SIZE));
  tgp->t_samplerp = &fgn_sampler;

  return (tgp);
} /* fgn_reader() */

/*
 * External sampler.
 * Stateless in packet mode, keeps state info in interval mode.
 * Reads the TRUs in input from a file or a pipe.
 *
 * Syntax is: external [type] [scale] ifile|ipipe [burst]
 *	where <ifile> is the file name where the TRUs are stored, <ipipe>
 *	is the command used to launch the program whose output is piped
 *	to the generator.
 *	<type> can be one of "packet", meaning that the number of TRUS
 *	per frame are contained in the file (one integer per line), or
 *	"interval", meaning that the time intervals between each packet
 *	and its successor (one floating point number per line)
 *	are in the file.  In interval mode, <scale> is a floating
 *	point factor (default 1) which is used to divide the intervals
 *	read from the file.
 *
 *	If a filename ends in .gz then the filter `gzip -cd' is used
 *	to read it.  If it ends in .Z then the filter `compress -cd'
 *	is used.  If it ends in .z then `zcat' is used.  Each line in
 *	the (decompressed) file contains an integer number and
 *	possibly a string followed by a newline.  Only the integer
 *	number is significant.
 */
static int
external_sampler (tg_args *argsp)
{
  struct file_args *tap = &argsp->file;
  int trus;

  if (tap->interval)
    {
      float interval;
      float last_time = tap->last_time;
      float scale = tap->scale;
      float limit = frametime * scale;

      trus = 0;
      while (last_time < limit)
	{
	  if (fscanf (tap->stream, "%g[^\n]\n", &interval) != 1)
	    break;
	  last_time += interval;
	  trus += 1;
	}
      tap->last_time = last_time - limit;
    }
  else
    if (fscanf (tap->stream, "%d[^\n]\n", &trus) != 1)
      trus = 0;

  return (trus * tap->burst);
} /* external_sampler() */

tg *
external_reader (char *lp)
{
  char *mode_type;
  FILE *pipe_stream;

  char *errstr;
  tg *tgp = allocate (sizeof (tg));
  struct file_args *tap = &tgp->t_args.file;
  arg_type args[] =
    {
      { "type", string_t, required },
      { "scale", float_t, optional },
      { "ifile", ifile_t, optional },
      { "burst", uns_t, optional },
      { "ipipe", ipipe_t, optional },
      { NULL }
    };

  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);

  get_arg ("type", mode_type);
  get_arg ("scale", tap->scale);
  get_arg ("ifile", tap->stream);
  get_arg ("burst", tap->burst);
  get_arg ("ipipe", pipe_stream);

  if (tap->burst == 0)
   tap->burst = 1;

  if (mode_type != NULL)
    {
      if (streq (mode_type, "interval"))
        {
	  if (tap->scale == 0)
	    tap->scale = 1;
          else if (tap->scale <= 0)
            return_error ("scale must be positive");
          tap->interval = TRUE;
          tap->last_time = 0;
        }
      else if (streq (mode_type, "packet"))
	{
	  if (tap->scale != 0)
	    return_error ("scale may not be specified with type `packet'");
	  tap->interval = FALSE;
	}
      else
	return_error ("type must be `packet' or `interval'");
    }

  if (tap->stream == NULL)
    tap->stream = pipe_stream;
  else if (pipe_stream != NULL)
    return_error ("only one of `file' and `pipe' must be specified");
  if (tap->stream == NULL)
    return_error ("one of `file' and `pipe' must be specified");
  tgp->t_samplerp = &external_sampler;

  return (tgp);
} /* external_reader() */

/*
 * Markov sampler
 * Keeps state info.
 * Generate TRUs according to a Markov chain.
 *  The chain is described in a text input file.  No comment lines or blank
 * lines are allowed.  Each state has a corresponding associated mean
 * throughput (called the state throughput in the following description),
 * while the effective throughput generated while in a state is a random
 * number having a constant distribution in the bin size of the throughputs.
 *	
 * The first line of the input file has the following format:
 *      steps=%u mint=%f maxt=%f binsize=%f\n
 * and contains:
 *  - A positive integer which is the number of possible different
 *    throughputs of the markov chain (currently unused).
 *  - A float which is the minimum throughput (currently unused).
 *  - A float which is the maximum throughput (currently unused).
 *  - A float which is the width of each throughput bin.
 *
 * The second line is currently ignored.
 *
 * Each subsequent line describes a state of the Markov chain.  The fields
 * in a line are numbers with the following meaning:
 * <1>	The state name, which is an integer number.
 * <2>	The probability of being the initial state.
 * <3>	The steady state probability (currently unused).
 * <4>  The throughput of the generator while in this state.  The
 *	throughputs must be normalized, that is, the sum for all the states
 *	of the throughput times the steady state probability must be 1.
 * <n p>... Any number (at least 1) of states that are reached from
 *	    this state, where N is the state name and P the
 *	    probability of switching to that state.  The sum of the
 *	    probabilities must be 1.
 *
 * Sintax is: markov traffic|tfactor file steplen [stepno] [seed] [burst]
 *	where <file> is the name of the file describing the chain, whose
 *	contents are described above. <steplen> is the float interval
 *	between state changes in TMU.  <stepno> is the positive integer
 *	number of steps after which the markov chain is reinitialised.
 *	This may make sense when the markov chain has transient states,
 *	that is states whose steady state probability is null, and one of
 *	them is chosen as the starting state.  When <stepno> is set to 0
 *	(default), the chain is never restarted, and transient states, once
 * abandoned, are never reentered.  */
void
init_markov_state (struct markov_args *tap)
{
  struct ma_state *sp;
  double cuminiprob;

  do
    {
      float r = frandom ();
      for (sp = tap->chain, cuminiprob = 0;
	   r < (cuminiprob += sp->iniprob) && sp->nextstate != NULL;
	   sp = sp->nextstate)
	continue;
    }
  while (sp->iniprob == 0);
  tap->cur_step = 1;
  tap->cur_sp = sp;
  tap->cur_thput = sp->thput + (frandom () - 0.5) * tap->binsize;
}

void
advance_markov_state (struct markov_args *tap)
{
  struct ma_edge *ep;
  float r = frandom ();

  if (tap->cur_step >= tap->stepno)
    init_markov_state (tap);
  else
    {
      for (ep = tap->cur_sp->ehead;
	   r < ep->cumjmpprob && ep->nextedge != NULL;
	   ep = ep->nextedge)
	continue;
      tap->cur_sp = ep->stp;
      tap->cur_thput = tap->cur_sp->thput + (frandom () - 0.5) * tap->binsize;
    }
}

static int
markov_sampler (tg_args *argsp)
{
  struct markov_args *tap = &argsp->markov;
  double trus = 0;
  double frame_end = curtime + frametime;
  double interval_start = curtime;
  double interval_end = tap->next_switch;

  setstate (&tap->rs);

  for (;;)
    {
      if (interval_end > frame_end)
	{
	  trus += tap->cur_thput * (frame_end - interval_start);
	  break;
	}
      trus += tap->cur_thput * (interval_end - interval_start);
      advance_markov_state (tap);
      interval_start = interval_end;
      interval_end += tap->steplen;
    }
  tap->next_switch = interval_end;

  return (floor (trus+0.5) * tap->burst);
} /* markov_sampler() */

tg *
markov_reader (char *lp)
{
  float traffic, tfactor, steplen, binsize;
  FILE *chainfile;
  int stepno;
  int nread, ma_lineno;
  double cuminiprob;
  struct ma_state *sp, *sp1;	/* used when building the markov chain */
  struct ma_edge *ep;		/* used when building the markov chain */
  unsigned int st, s1;		/* used for fscanf */
  float t, p, p1;		/* used for fscanf */
  int seed;

  char *errstr;
  tg *tgp = allocate (sizeof (tg));
  struct markov_args *tap = &tgp->t_args.markov;
  arg_type args[] =
  {
    { "traffic", float_t, optional },
    { "tfactor", float_t, optional },
    { "file", ifile_t, required },
    { "steplen", float_t, required },
    { "seed", int_t, optional },
    { "burst", uns_t, optional },
    { "stepno", int_t, optional },
    { NULL }
  };
 
  errstr = lineparser (lp, args);
  if (errstr != NULL)
    return_error (errstr);

  get_arg ("traffic", traffic);
  get_arg ("tfactor", tfactor);
  get_arg ("file", chainfile);
  get_arg ("steplen", steplen);
  get_arg ("seed", seed);
  get_arg ("stepno", stepno);

  if (seed == 0)
    seed = random ();
  
  get_arg ("burst", tap->burst);
  if (tap->burst == 0)
    tap->burst = 1;

  if (steplen <= 0)
      return_error ("steplen must be positive");

  if (traffic == 0)
    {
      if (tfactor == 0)
	{
	  input_warning ("generator has null traffic");
	  tgp->t_samplerp = &null_sampler;
	  return (tgp);
	}
      if (tfactor < 0)
	return_error ("tfactor must be positive");
      if (reference_traffic == 0)
	return_error ("tfactor requires ref_traffic to be specified");
      traffic = tfactor * reference_traffic;
    }
  else if (traffic < 0)
    return_error ("traffic must be positive");
  else if (tfactor != 0)
    return_error ("only one of traffic and tfactor can be specified");

  tap->stepno = stepno; 
  tap->steplen = steplen;
  tap->next_switch = steplen;

  /* Read the first line, getting the size of the bin
     of the relative throughput. */
  if (fscanf (chainfile, " steps=%*u mint=%*f maxt=%*f binsize=%f%*[^\n]",
	      &binsize) != 1)
    return_error ("invalid step while reading line 1 from Markov input file");
  if (binsize <= 0)
    return_error ("the bin size must be positive (line 1 of Markov input file)");
  tap->binsize = binsize * traffic;

  /* Read and discard the first newline and the entire second line. */
  fscanf (chainfile, "\n%*[^\n]\n");
 
  /*
   * Building the graph of the Markov chain using states and edges in
   * a linked list of states.  Each state has a linked list of edges.
   * After building the chain, each edge is linked to its destination state.
   */
  tap->chain = NULL;
  cuminiprob = 0;
  for (ma_lineno = 3;
       fscanf (chainfile, " %u%*[ \t]%f%*[ \t]%*f%*[ \t]%f", &st, &p, &t)
	 == 3;
       ma_lineno++)
    {
      double cumprob;

      sp = allocate (sizeof (struct ma_state));
      sp->state = st;
      if (p < 0 || p > 1)
	{ 
	  char errbuf [200];
	  sprintf (errbuf,
		   "invalid probability %g while reading line %d from Markov input file",
		   p, ma_lineno);
	  return_error (errbuf);
	}
      cuminiprob += p;
      sp->iniprob = p;
      sp->thput = t * traffic;
      sp->ehead = NULL;
      sp->nextstate = tap->chain;
      tap->chain = sp;

      for (cumprob = 0;;)
	{ 
	  nread = fscanf (chainfile, "%*[ \t]%u%*[ \t]%f", &s1, &p1);
	  if (nread == 0 || nread == EOF)
	    break;
	  if (nread == 1)
	    { 
	      char errbuf [200];
	      sprintf (errbuf,
		       "error while reading line %d from Markov input file",
		       ma_lineno);
	      return_error (errbuf);
	    }
	  /* nread == 2 */
	  if (p1 < 0 || p1 > 1)
	    { 
	      char errbuf [200];
	      sprintf (errbuf,
		       "invalid probability %g while reading line %d from Markov input file",
		       p1, ma_lineno);
	      return_error (errbuf);
	    }
	  ep = allocate (sizeof (struct ma_edge));
	  ep->st = s1;
	  /* The list of the edges is built by head insertion, and the
	     cumulative probability of taking that edge is computed
	     during the build, so the first element has the highest
	     cumulative probability. */
	  ep->cumjmpprob = cumprob;
	  cumprob += p1;
       	  ep->nextedge = sp->ehead;
	  sp->ehead = ep;
	}
      if (fabs (cumprob - 1) > 5e-4)
	{ 
	  char errbuf [200];
	  sprintf (errbuf,
		   "sum of probabilities not 1 at line %d from Markov input file",
		   ma_lineno);
	  return_error (errbuf);
	}
    }
  if (feof (chainfile))
    fclose (chainfile);
  else
    { 
      char errbuf [200];
      sprintf (errbuf,
	       "error while reading line %d from Markov input file",
	       ma_lineno);
      return_error (errbuf);
    }
  if (fabs (cuminiprob - 1) > 1e-4)
    return_error ("sum of state probabilities in input file is not 1");
  for (sp = tap->chain; sp != NULL; sp = sp->nextstate)
    for (ep = sp->ehead; ep != NULL; ep = ep->nextedge)
      {
	for (sp1 = tap->chain; sp1 != NULL; sp1 = sp1->nextstate)
	  if (sp1->state == ep->st)
	    {
	      ep->stp = sp1;
	      break;
	    }
	if (sp1 == NULL)
	  { 
	    char errbuf [200];
	    sprintf (errbuf,
		     "state #%d reached from state #%d does not exist",
		     ep->st, sp->state);
	    return_error (errbuf);
	  }
      }


  /* Compute initial state for the Markov chain. */
  setstate (initstate (seed, &tap->rs, RND_STATE_SIZE));
  init_markov_state (tap);
  tgp->t_samplerp = &markov_sampler;

  return (tgp);
} /* markov_reader() */
