/* The stcollect package provides a simple interface for programs
   that need to compute simple statistics on scalar data.  See the
   function for details. */

/*
 *  Function definitions for the stcollect package:
 *  st_init	initialise a statcollect structure
 *  st_collect	collect n samples
 *  st_add	add two st_data structures
 *  st_n	get the number of collected samples
 *  st_min	get the minimum collected sample
 *  st_max	get the maximum collected sample
 *  st_avg	get the average from the collected samples
 *  st_var	get the variance from the collected samples
 *  st_stddev	get the standard deviation from the collected samples
 */

#include <limits.h>
#include <math.h>
#include "stcollect.h"
#undef NULL
#define NULL 0


/*
 * Allocate space for a st_data structure with calloc if argument
 * is NULL, else clear the pointed to structure.  Return the pointer
 * to the structure, or NULL if calloc failed.
 */
st_data *
st_init ARG1 (st_data *statp)
{
  if (statp == NULL)
    {
      statp = (st_data *) calloc (1, sizeof (st_data));
      if (statp == NULL)
	return (NULL);
    }
  else
    {
      memset ((char *)statp, 0, sizeof (st_data));
    }
     
  return (statp);
} /* st_init */

void
st_collect ARG3 (st_data *statp, double x, int n)
{
  if (statp->samples == 0)
    {
      statp->max = x;
      statp->min = x;
    }
  else
    {
      if (x > statp->max)
	statp->max = x;
      else if (x < statp->min)
	statp->min = x;
    }
  if (n == 1)
    {
      statp->samples += 1;
      statp->sum += x;
      statp->sumsqr += x*x;
    }
  else
    {
      double sum = n*x;
      statp->samples += n;
      statp->sum += sum;
      statp->sumsqr += sum*x;
    }
} /* st_collect() */

void
st_add ARG2 (st_data *accumulator, st_data *addendum)
{
  accumulator->samples += addendum->samples;
  accumulator->sum += addendum->sum;
  accumulator->sumsqr += addendum->sumsqr;
  if (addendum->min < accumulator->min)
    accumulator->min = addendum->min;
  if (addendum->max > accumulator->max)
    accumulator->max = addendum->max;
} /* st_add() */

double
st_n ARG1 (st_data *statp)
{
  return (statp->samples);
} /* st_n() */

double
st_min ARG1 (st_data *statp)
{
  return (statp->min);
} /* st_min() */

double
st_max ARG1 (st_data *statp)
{
  return (statp->max);
} /* st_max() */

double
st_avg ARG1 (st_data *statp)
{
  return (statp->sum / statp->samples);
} /* st_avg() */

double
st_var ARG1 (st_data *statp)
{
  double var = ((statp->sumsqr - statp->sum * statp->sum / statp->samples) /
		(statp->samples - 1));
  if (var > 0)
    return (var);
  else
    return (0);
} /* st_var() */

double
st_stddev ARG1 (st_data *statp)
{
  return sqrt (st_var (statp));
} /* st_stddev() */

