#ifdef DEBUG
# define allocate(n) trace_calloc(__FILE__, __LINE__, n, 1)
#endif
/*
 * This function builds a character graph displayable on a tty terminal.
 * CS[COLS] contains the points to be plotted.  Upon exit,
 * SCREEN[ROWS+1] contains pointers to strings long COLS+2, printable with
 *   for (row = rows; row > 0; row--) printf(screen[row]);
 * ROWS and COLS are the graph size, a nonzero FILL means that the lower
 * part of the screen should be filled with asterisks.
 */
#define TABLEN 8
#define TAB '\t'
#define EOL '\n'
#define EOS '\0'
#define SPC ' '
#define STAR '*'
#define DOT '.'
#define HPN '-'
#define UDS '_'
void
ttygraph (float cs[], int rows, int cols, char *screen[], int fill)
{
  float span, min, max;
  register int row, col;
  register int tabcol, dstcol, srccol;
  register char *cp;

  min = max = cs[0];
  for (col = 1; col < cols; col++) /* find min and max */
    if (cs[col] < min)
      min = cs[col];
    else if (cs[col] > max)
      max = cs[col];
  span = max - min;
  if (span == 0)
    for (col = 0; col < cols; col++) /* a line in the middle of the screen */
      cs[col] = rows;
  else
    for (col = 0; col < cols; col++) /* normalise categories */
      if (fill)
	cs[col] = 2*rows * (cs[col]-min)/span;
      else
	cs[col] = (2*rows-1) * (cs[col]-min)/span + 1;

  for (row = rows; row > 0; row--) /* fill screen */
    {
      char *line = screen[row];
      for (col = 0; col < cols; col++) /* put symbols for graph */
	if (fill || cs[col] < 2*row+1)
	  {
	    if (cs[col] >= 2*row)
	      line[col] = (fill) ? STAR : HPN;
	    else if (cs[col] >= 2*row-1)
	      line[col] = (fill) ? DOT : UDS;
	    else
	      line[col] = (fill && row==1) ? UDS : SPC;
	  }
	else
	  line[col] = SPC;

      /* Printout optimisations. */
      tabcol = TABLEN;
      dstcol = 0;
      srccol = 0;
      while (srccol < cols)
	{
	  if (srccol >= tabcol)
	    tabcol += TABLEN;
	  if (line[srccol] == SPC)
	    {
	      for (cp = &line[srccol+1]; *cp == SPC && cp < &line[cols]; cp++)
		continue;
	      if (cp == &line[cols])
		break;
	      if (cp - line >= tabcol)
		{
		  line[dstcol++] = TAB;
		  srccol = tabcol;
		  continue;
		}
	    }
	  line[dstcol++] = line[srccol++];
	}
      line[dstcol] = EOL;
      line[dstcol+1] = EOS;
    }
} /* ttygraph() */


/*
 * This function fills the CS[CATN] with floats representing the
 * number of samples in Q[SAMPLES] belonging to each of CATN categories.
 * MIN and MAX are the maximum and minimum values of the numbers in Q[].
 */
void
categorise (float q[], int samples, float min, float max,
	    int catn, float cs[])
{
  float step;
  register int i, c;

  step = (max - min) / catn;

  for (c = 0; c < catn; c++)
    cs[c] = 0;
  if (step <= 0)		/* max == min */
    step = 1;			/* this way c will always be 0 */
  for (i = 0; i < samples; i++)
    {
      c = (q[i]-min) / step;
      if (c >= 0 && c < catn)
	cs[c] += 1;
      else if (c < 0)
	cs[0] += 1;
      else
	cs[catn-1] += 1;
    }
}


/*
 * This function builds a character graph displayable on a tty terminal.
 * Q[SAMPLES] contains the samples whose distribution is to be plotted.
 * MIN, MAX, AVG, SIGMA are the statistics of the samples.
 * Upon exit, SCREEN[ROWS+1][COLS+2] is filled with chars, printable with
 *   for (row = rows; row > 0; row--) printf(screen[row]);
 * ROWS and COLS are the graph size.
 */
#define ABS(a)	(((a) < 0) ? -(a) : (a))
void
ttydistrib (float q[], int samples,
	    float min, float max, float avg, float sigma,
	    int rows, int cols, char *screen[])
{
  float *cs;
  int i, col;

  cs = allocate (cols * sizeof (float));
  categorise (q, samples, min, max, cols, cs);

  ttygraph (cs, rows, cols, screen, 1);

  free (cs);

  /* +/-3sigma marks */
  for (i = -3; i <= 3; i++)
    {
      col = (avg - min + i*sigma) / (max-min) * cols;
      if (col >= 0 && col < cols)
	screen[rows][col] = '0' + ABS(i);
    }
} /* ttydistrib() */
